# -*- coding:utf8 -*-
# @TIME    : 2022/7/
# @Author  : Cheng Guo
# @File    :

import math
import numpy as np
from QuICT.core.gate import *
from QuICT.core.gate.backend.mct.mct_linear_dirty_aux import MCTLinearHalfDirtyAux
from QuICT.core.gate.backend.mct.mct_one_aux import MCTOneAux
from QuICT.qcda.optimization.commutative_optimization import *


class CNFSATOracle:
    def __init__(self, simu=None):
        self.simulator = simu

    def circuit(self) -> CompositeGate:
        return self._cgate

    def run(
        self, cnf_file: str, ancilla_qubits_num: int = 3, dirty_ancilla: int = 0
    ) -> int:
        """Run CNF algorithm

        Args:
            cnf_file (str): The file path
            ancilla_qubits_num (int): >= 3
            dirty_ancilla (int): 0 for clean and >0 for dirty
        """
        # check if Aux > 2
        assert ancilla_qubits_num > 2, "Need at least 3 auxiliary qubit."

        # Step 1: Read CNF File
        variable_number, clause_number, CNF_data = self.read_CNF(cnf_file)

        # Step 2: Construct Circuit
        self._cgate = CompositeGate()
        p = math.floor((ancilla_qubits_num + 1) / 2)
        depth = math.ceil(math.log(clause_number, p))
        target = variable_number
        if clause_number == 1:
            controls = CNF_data[1]
            controls_abs = []
            controls_X = []
            current_Aux = target + 1
            for i in range(len(controls)):
                if controls[i] < 0:
                    controls_abs.append(-controls[i] - 1)
                if controls[i] > 0:
                    controls_abs.append(controls[i] - 1)
                    controls_X.append(controls[i] - 1)
            for i in range(len(controls_X)):
                X | self._cgate(controls_X[i])
            X | self._cgate(target)
            if controls_abs != []:
                MCTOneAux().execute(len(controls_abs) + 2) | self._cgate(
                    controls_abs + [target, current_Aux]
                )

            for i in range(len(controls_X)):
                X | self._cgate(controls_X[i])
        else:
            if dirty_ancilla == 0:
                if clause_number < p + 1:
                    controls_abs = []
                    for j in range(clause_number):
                        controls_abs.append(variable_number + j + 1)
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            j + 1,
                            j + 1,
                            variable_number + j + 1,
                            depth - 1,
                            depth,
                        )
                    if controls_abs != []:
                        MCTOneAux().execute(len(controls_abs) + 2) | self._cgate(
                            controls_abs + [target, target - 1]
                        )
                    else:
                        X | self._cgate(target)
                    for j in range(clause_number):
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            j + 1,
                            j + 1,
                            variable_number + j + 1,
                            depth - 1,
                            depth,
                        )
                else:
                    block_len = math.ceil(clause_number / p)
                    block_number = math.ceil(clause_number / block_len)
                    controls = []

                    for j in range(block_number):
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            j * block_len + 1,
                            np.minimum((j + 1) * block_len, clause_number),
                            variable_number + ancilla_qubits_num - p + 1 + j,
                            depth - 1,
                            depth,
                        )
                        controls.append(
                            variable_number + ancilla_qubits_num - p + 1 + j
                        )

                    current_Aux = variable_number + 1
                    if controls != []:
                        MCTOneAux().execute(len(controls) + 2) | self._cgate(
                            controls + [target, current_Aux]
                        )

                    for j in range(block_number):
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            j * block_len + 1,
                            np.minimum((j + 1) * block_len, clause_number),
                            variable_number + ancilla_qubits_num - p + 1 + j,
                            depth - 1,
                            depth,
                        )
            else:  # dirty_ancilla =1
                if clause_number < p + 1:
                    controls_abs = []
                    for j in range(clause_number):
                        controls_abs.append(variable_number + j + 1)
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            j + 1,
                            j + 1,
                            variable_number + j + 1,
                            depth - 1,
                            depth,
                        )
                    if controls_abs != []:
                        MCTOneAux().execute(len(controls_abs) + 2) | self._cgate(
                            controls_abs + [target, target - 1]
                        )

                    for j in range(clause_number):
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            j + 1,
                            j + 1,
                            variable_number + j + 1,
                            depth - 1,
                            depth,
                        )
                else:
                    if clause_number == 2:  # to do

                        CCX | self._cgate(
                            [
                                variable_number + ancilla_qubits_num - 1,
                                variable_number + ancilla_qubits_num,
                                target,
                            ]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            1,
                            1,
                            variable_number + ancilla_qubits_num - 1,
                            depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [
                                variable_number + ancilla_qubits_num - 1,
                                variable_number + ancilla_qubits_num,
                                target,
                            ]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            clause_number,
                            clause_number,
                            variable_number + ancilla_qubits_num,
                            depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [
                                variable_number + ancilla_qubits_num - 1,
                                variable_number + ancilla_qubits_num,
                                target,
                            ]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            1,
                            1,
                            variable_number + ancilla_qubits_num - 1,
                            depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [
                                variable_number + ancilla_qubits_num - 1,
                                variable_number + ancilla_qubits_num,
                                target,
                            ]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            ancilla_qubits_num,
                            clause_number,
                            clause_number,
                            variable_number + ancilla_qubits_num,
                            depth - 1,
                            depth,
                        )

                    else:
                        block_len = math.ceil(clause_number / p)
                        block_number = math.ceil(clause_number / block_len)
                        current_depth = depth
                        # even
                        if block_number == 2:
                            CCX | self._cgate(
                                [
                                    variable_number + ancilla_qubits_num - 1,
                                    variable_number + ancilla_qubits_num,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1,
                                block_len,
                                variable_number + ancilla_qubits_num - 1,
                                depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    variable_number + ancilla_qubits_num - 1,
                                    variable_number + ancilla_qubits_num,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                block_len + 1,
                                clause_number,
                                variable_number + ancilla_qubits_num,
                                depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    variable_number + ancilla_qubits_num - 1,
                                    variable_number + ancilla_qubits_num,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1,
                                block_len,
                                variable_number + ancilla_qubits_num - 1,
                                depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    variable_number + ancilla_qubits_num - 1,
                                    variable_number + ancilla_qubits_num,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                block_len + 1,
                                clause_number,
                                variable_number + ancilla_qubits_num,
                                depth - 1,
                                depth,
                            )

                        else:  # block_number > 2
                            c = []
                            for i in range(
                                variable_number,
                                variable_number + ancilla_qubits_num + 1,
                            ):
                                if i != target:
                                    c.append(i)

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - block_number],
                                    c[ancilla_qubits_num - 2 * (block_number - 1)],
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1,
                                block_len,
                                c[ancilla_qubits_num - block_number],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - block_number],
                                    c[ancilla_qubits_num - 2 * (block_number - 1)],
                                    target,
                                ]
                            )

                            # Control bit variable_ Number+Aux - (block_number-1)+2 - j
                            # put a clause, another control bit
                            # (will be the same as the previous target in this for)
                            # and the target will rise in turn
                            for j in range(1, block_number - 2):
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    ancilla_qubits_num,
                                    1 + j * block_len,
                                    (1 + j) * block_len,
                                    c[ancilla_qubits_num - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )

                                # topPhase
                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1 + (block_number - 2) * block_len,
                                (block_number - 1) * block_len,
                                c[ancilla_qubits_num - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1 + (block_number - 1) * block_len,
                                clause_number,
                                c[ancilla_qubits_num - 1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1 + (block_number - 2) * block_len,
                                (block_number - 1) * block_len,
                                c[ancilla_qubits_num - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )

                            for j in range(block_number - 3, 0, -1):
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    ancilla_qubits_num,
                                    1 + j * block_len,
                                    (1 + j) * block_len,
                                    c[ancilla_qubits_num - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - block_number],
                                    c[ancilla_qubits_num - 2 * (block_number - 1)],
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1,
                                block_len,
                                c[ancilla_qubits_num - block_number],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - block_number],
                                    c[ancilla_qubits_num - 2 * (block_number - 1)],
                                    target,
                                ]
                            )

                            # repeat
                            for j in range(1, block_number - 2):
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    ancilla_qubits_num,
                                    1 + j * block_len,
                                    (1 + j) * block_len,
                                    c[ancilla_qubits_num - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )

                                # topPhase
                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1 + (block_number - 2) * block_len,
                                (block_number - 1) * block_len,
                                c[ancilla_qubits_num - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1 + (block_number - 1) * block_len,
                                clause_number,
                                c[ancilla_qubits_num - 1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                ancilla_qubits_num,
                                1 + (block_number - 2) * block_len,
                                (block_number - 1) * block_len,
                                c[ancilla_qubits_num - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[ancilla_qubits_num - 2],
                                    c[ancilla_qubits_num - 1],
                                    c[ancilla_qubits_num - 2 - (block_number - 1)],
                                ]
                            )

                            for j in range(block_number - 3, 0, -1):
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    ancilla_qubits_num,
                                    1 + j * block_len,
                                    (1 + j) * block_len,
                                    c[ancilla_qubits_num - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[
                                            ancilla_qubits_num
                                            - (block_number - 1)
                                            + j
                                            - 1
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            + j
                                        ],
                                        c[
                                            ancilla_qubits_num
                                            - 2 * (block_number - 1)
                                            - 1
                                            + j
                                        ],
                                    ]
                                )

    def read_CNF(self, cnf_file):
        # file analysis

        variable_number = 0
        clause_number = 0
        CNF_data = []
        f = open(cnf_file, "r")
        for line in f.readlines():
            new = line.strip().split()
            int_new = []
            if new[0] == "p":
                variable_number = int(new[2])
                clause_number = int(new[3])
            else:
                for x in new:
                    if (x != "0") and (int(x) not in int_new):
                        int_new.append(int(x))
                        if (-int(x)) in int_new:
                            int_new = []
                            break

            CNF_data.append(int_new)

        f.close()
        return variable_number, clause_number, CNF_data

    def clause(
        self,
        CNF_data,
        variable_number: int,
        Aux: int,
        StartID: int,
        EndID: int,
        target: int,
        current_depth: int,
        depth: int,
    ):

        p = math.floor((Aux + 1) / 2)
        if StartID == EndID:  # n= variable_number + Aux + 1
            controls = CNF_data[StartID]
            controls_abs = []
            controls_X = []
            for i in range(len(controls)):
                if controls[i] < 0:
                    controls_abs.append(-controls[i] - 1)
                if controls[i] > 0:
                    controls_abs.append(controls[i] - 1)
                    controls_X.append(controls[i] - 1)
            for i in range(len(controls_X)):
                X | self._cgate(controls_X[i])
            X | self._cgate(target)

            d = set(range(variable_number + 1 + Aux))
            d.remove(target)
            for j in controls_abs:
                d.remove(j)
            d = list(d)
            if controls_abs != []:
                MCTLinearHalfDirtyAux().execute(
                    len(controls_abs), (1 + variable_number + Aux)
                ) | self._cgate(controls_abs + d + [target])

            for i in range(len(controls_X)):
                X | self._cgate(controls_X[i])
        else:  # StartID != EndID
            if (EndID - StartID) == 1:  # StartID +1 == EndID
                if ((depth - current_depth) % 2) == 1:  # odd

                    CCX | self._cgate([variable_number + 1, variable_number, target])
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        StartID,
                        StartID,
                        variable_number + 1,
                        current_depth - 1,
                        depth,
                    )

                    CCX | self._cgate([variable_number + 1, variable_number, target])
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        EndID,
                        EndID,
                        variable_number,
                        current_depth - 1,
                        depth,
                    )

                    CCX | self._cgate([variable_number + 1, variable_number, target])
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        StartID,
                        StartID,
                        variable_number + 1,
                        current_depth - 1,
                        depth,
                    )

                    CCX | self._cgate([variable_number + 1, variable_number, target])
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        EndID,
                        EndID,
                        variable_number,
                        current_depth - 1,
                        depth,
                    )

                # even    target in {variable_number + ancilla_qubits_num - p + 1 + j}
                else:

                    CCX | self._cgate(
                        [variable_number + Aux - 1, variable_number + Aux, target]
                    )
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        StartID,
                        StartID,
                        variable_number + Aux - 1,
                        current_depth - 1,
                        depth,
                    )

                    CCX | self._cgate(
                        [variable_number + Aux - 1, variable_number + Aux, target]
                    )
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        EndID,
                        EndID,
                        variable_number + Aux,
                        current_depth - 1,
                        depth,
                    )

                    CCX | self._cgate(
                        [variable_number + Aux - 1, variable_number + Aux, target]
                    )
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        StartID,
                        StartID,
                        variable_number + Aux - 1,
                        current_depth - 1,
                        depth,
                    )

                    CCX | self._cgate(
                        [variable_number + Aux - 1, variable_number + Aux, target]
                    )
                    self.clause(
                        CNF_data,
                        variable_number,
                        Aux,
                        EndID,
                        EndID,
                        variable_number + Aux,
                        current_depth - 1,
                        depth,
                    )

            else:  # EndID - StartID > 1
                if (
                    EndID - StartID
                ) < p:  # if block_number == 1 and EndID - StartID > 1

                    c = []
                    for i in range(variable_number, variable_number + Aux + 1):
                        if i != target:
                            c.append(i)

                    if ((depth - current_depth) % 2) == 1:

                        CCX | self._cgate(
                            [c[EndID - StartID], c[2 * (EndID - StartID) - 1], target]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            StartID,
                            StartID,
                            c[EndID - StartID],
                            current_depth - 1,
                            depth,
                        )
                        CCX | self._cgate(
                            [c[EndID - StartID], c[2 * (EndID - StartID) - 1], target]
                        )

                        for j in range(1, EndID - StartID - 1):
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[(EndID - StartID) - j],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )

                            # topPhase
                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[1],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            EndID,
                            EndID,
                            c[0],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[1],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])

                        # downPhase
                        for j in range(EndID - StartID - 2, 0, -1):
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[(EndID - StartID) - j],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )

                        CCX | self._cgate(
                            [c[EndID - StartID], c[2 * (EndID - StartID) - 1], target]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            StartID,
                            StartID,
                            c[EndID - StartID],
                            current_depth - 1,
                            depth,
                        )
                        CCX | self._cgate(
                            [c[EndID - StartID], c[2 * (EndID - StartID) - 1], target]
                        )

                        # repeat....

                        # Restore each location
                        # Control bit variable_ Number+Aux - (EndID StartID)+2 - j put a clause, another control bit
                        # (will be the same as the previous target in this for) and the target will rise in turn
                        for j in range(1, EndID - StartID - 1):
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[(EndID - StartID) - j],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )

                            # topPhase
                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[1],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            EndID,
                            EndID,
                            c[0],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[1],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate([c[1], c[0], c[EndID - StartID + 1]])

                        # downPhase
                        for j in range(EndID - StartID - 2, 0, -1):
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[(EndID - StartID) - j],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[(EndID - StartID) - j],
                                    c[2 * (EndID - StartID) - 1 - j],
                                    c[2 * (EndID - StartID) - j],
                                ]
                            )

                    else:  # even
                        CCX | self._cgate(
                            [
                                c[Aux - 1 - EndID + StartID],
                                c[Aux - 2 * (EndID - StartID)],
                                target,
                            ]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            StartID,
                            StartID,
                            c[Aux - 1 - EndID + StartID],
                            current_depth - 1,
                            depth,
                        )
                        CCX | self._cgate(
                            [
                                c[Aux - 1 - EndID + StartID],
                                c[Aux - 2 * (EndID - StartID)],
                                target,
                            ]
                        )

                        for j in range(1, EndID - StartID - 1):
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[Aux - (EndID - StartID) + j - 1],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )

                            # topPhase
                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[Aux - 2],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            EndID,
                            EndID,
                            c[Aux - 1],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[Aux - 2],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )

                        for j in range(EndID - StartID - 2, 0, -1):
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[Aux - (EndID - StartID) + j - 1],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )

                        CCX | self._cgate(
                            [
                                c[Aux - 1 - EndID + StartID],
                                c[Aux - 2 * (EndID - StartID)],
                                target,
                            ]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            StartID,
                            StartID,
                            c[Aux - 1 - EndID + StartID],
                            current_depth - 1,
                            depth,
                        )
                        CCX | self._cgate(
                            [
                                c[Aux - 1 - EndID + StartID],
                                c[Aux - 2 * (EndID - StartID)],
                                target,
                            ]
                        )

                        for j in range(1, EndID - StartID - 1):
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[Aux - (EndID - StartID) + j - 1],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )

                            # topPhase
                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[Aux - 2],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            EndID,
                            EndID,
                            c[Aux - 1],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )
                        self.clause(
                            CNF_data,
                            variable_number,
                            Aux,
                            (EndID - 1),
                            EndID - 1,
                            c[Aux - 2],
                            current_depth - 1,
                            depth,
                        )

                        CCX | self._cgate(
                            [c[Aux - 2], c[Aux - 1], c[Aux - 2 - EndID + StartID]]
                        )

                        for j in range(EndID - StartID - 2, 0, -1):
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + j,
                                StartID + j,
                                c[Aux - (EndID - StartID) + j - 1],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[Aux - (EndID - StartID) + j - 1],
                                    c[Aux - 2 * (EndID - StartID) + j],
                                    c[Aux - 2 * (EndID - StartID) - 1 + j],
                                ]
                            )

                else:  # EndID-StartID > p-1  block number >1
                    block_len = math.ceil((EndID - StartID + 1) / p)
                    block_number = math.ceil((EndID - StartID + 1) / block_len)
                    if block_number == 2:
                        # odd  target :variable_number + ancilla_qubits_num - p + 1 + j
                        if ((depth - current_depth) % 2) == 1:

                            CCX | self._cgate(
                                [variable_number + 1, variable_number, target]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID + block_len - 1,
                                variable_number + 1,
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [variable_number + 1, variable_number, target]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + block_len,
                                EndID,
                                variable_number,
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [variable_number + 1, variable_number, target]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID + block_len - 1,
                                variable_number + 1,
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [variable_number + 1, variable_number, target]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + block_len,
                                EndID,
                                variable_number,
                                current_depth - 1,
                                depth,
                            )
                        else:  # 当前偶数层

                            CCX | self._cgate(
                                [
                                    variable_number + Aux - 1,
                                    variable_number + Aux,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID + block_len - 1,
                                variable_number + Aux - 1,
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    variable_number + Aux - 1,
                                    variable_number + Aux,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + block_len,
                                EndID,
                                variable_number + Aux,
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    variable_number + Aux - 1,
                                    variable_number + Aux,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID + block_len - 1,
                                variable_number + Aux - 1,
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    variable_number + Aux - 1,
                                    variable_number + Aux,
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + block_len,
                                EndID,
                                variable_number + Aux,
                                current_depth - 1,
                                depth,
                            )
                    else:
                        # block number >2
                        c = []
                        for i in range(variable_number, variable_number + Aux + 1):
                            if i != target:
                                c.append(i)
                        if ((depth - current_depth) % 2) == 1:

                            CCX | self._cgate(
                                [
                                    c[block_number - 1],
                                    c[2 * (block_number - 1) - 1],
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID + block_len - 1,
                                c[block_number - 1],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[block_number - 1],
                                    c[2 * (block_number - 1) - 1],
                                    target,
                                ]
                            )

                            # control bit variable_number + Aux - (block_number-1) + 2 -j
                            # put clause, another control bit and target accension in accordance
                            for j in range(1, block_number - 2):
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (j + 1) * block_len,
                                    c[(block_number - 1) - j],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )

                                # topPhase
                            CCX | self._cgate([c[1], c[0], c[block_number]])
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate([c[1], c[0], c[block_number]])
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 1) * block_len,
                                EndID,
                                c[0],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate([c[1], c[0], c[block_number]])
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate([c[1], c[0], c[block_number]])

                            # downPhase
                            for j in range(block_number - 1 - 2, 0, -1):
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (j + 1) * block_len,
                                    c[(block_number - 1) - j],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )

                            CCX | self._cgate(
                                [
                                    c[block_number - 1],
                                    c[2 * (block_number - 1) - 1],
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID + block_len - 1,
                                c[block_number - 1],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[block_number - 1],
                                    c[2 * (block_number - 1) - 1],
                                    target,
                                ]
                            )

                            # repeat....

                            # 还原各个位置
                            # 控制位variable_number + Aux - (block_number-1) + 2 -j 放 clause
                            # 另一个 控制位(将与此for内的前一个target 相同)并和target 依次上升
                            for j in range(1, block_number - 1 - 1):
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (j + 1) * block_len,
                                    c[(block_number - 1) - j],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )

                                # topPhase
                            CCX | self._cgate([c[1], c[0], c[block_number]])
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate([c[1], c[0], c[block_number]])
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 1) * block_len,
                                EndID,
                                c[0],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate([c[1], c[0], c[block_number]])
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate([c[1], c[0], c[block_number]])

                            # downPhase
                            for j in range(block_number - 1 - 2, 0, -1):
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (j + 1) * block_len,
                                    c[(block_number - 1) - j],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[(block_number - 1) - j],
                                        c[2 * (block_number - 1) - 1 - j],
                                        c[2 * (block_number - 1) - j],
                                    ]
                                )

                        else:  # even
                            CCX | self._cgate(
                                [
                                    c[Aux - block_number],
                                    c[Aux - 2 * (block_number - 1)],
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID - 1 + block_len,
                                c[Aux - block_number],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[Aux - block_number],
                                    c[Aux - 2 * (block_number - 1)],
                                    target,
                                ]
                            )

                            for j in range(1, block_number - 2):
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (1 + j) * block_len,
                                    c[Aux - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )

                                # topPhase
                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[Aux - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 1) * block_len,
                                EndID,
                                c[Aux - 1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[Aux - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )

                            for j in range(block_number - 3, 0, -1):
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (1 + j) * block_len,
                                    c[Aux - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )

                            CCX | self._cgate(
                                [
                                    c[Aux - block_number],
                                    c[Aux - 2 * (block_number - 1)],
                                    target,
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID,
                                StartID - 1 + block_len,
                                c[Aux - block_number],
                                current_depth - 1,
                                depth,
                            )
                            CCX | self._cgate(
                                [
                                    c[Aux - block_number],
                                    c[Aux - 2 * (block_number - 1)],
                                    target,
                                ]
                            )

                            # re
                            for j in range(1, block_number - 1 - 1):
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (1 + j) * block_len,
                                    c[Aux - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )

                                # topPhase
                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[Aux - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 1) * block_len,
                                EndID,
                                c[Aux - 1],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )
                            self.clause(
                                CNF_data,
                                variable_number,
                                Aux,
                                StartID + (block_number - 2) * block_len,
                                StartID + (block_number - 1) * block_len - 1,
                                c[Aux - 2],
                                current_depth - 1,
                                depth,
                            )

                            CCX | self._cgate(
                                [
                                    c[Aux - 2],
                                    c[Aux - 1],
                                    c[Aux - 2 - (block_number - 1)],
                                ]
                            )

                            for j in range(block_number - 3, 0, -1):
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )
                                self.clause(
                                    CNF_data,
                                    variable_number,
                                    Aux,
                                    StartID + j * block_len,
                                    StartID - 1 + (1 + j) * block_len,
                                    c[Aux - (block_number - 1) + j - 1],
                                    current_depth - 1,
                                    depth,
                                )
                                CCX | self._cgate(
                                    [
                                        c[Aux - (block_number - 1) + j - 1],
                                        c[Aux - 2 * (block_number - 1) + j],
                                        c[Aux - 2 * (block_number - 1) - 1 + j],
                                    ]
                                )



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/11/12 上午10:10
# @Author  : Kaiqi Li
# @File    : proxy_unit_test.py\

import os
import unittest
from concurrent.futures import ProcessPoolExecutor, as_completed
import numpy as np

import cupy as cp
from cupy import cuda
from cupy.cuda import nccl

from QuICT.utility import Proxy


def sender(ndev, uid, dev_id):
    proxy = Proxy(ndev, uid, dev_id)
    based_data = np.random.rand(10).astype(np.complex64)
    gpu_data = cp.array(based_data)

    proxy.send(gpu_data, 1)

    return based_data


def receiver(ndev, uid, dev_id):
    proxy = Proxy(ndev, uid, dev_id)
    gpu_receive_buff = cp.zeros(10, dtype=np.complex64)

    proxy.recv(gpu_receive_buff, 0)

    return gpu_receive_buff.get()


def broadcast(ndev, uid, dev_id):
    proxy = Proxy(ndevs=ndev, uid=uid, dev_id=dev_id)
    if dev_id == 0:
        based_data = np.random.rand(10).astype(np.complex64)
        gpu_bd = cp.array(based_data)
    else:
        gpu_bd = cp.zeros(10, dtype=cp.complex64)

    proxy.broadcast(gpu_bd, 0)

    return gpu_bd.get()


@unittest.skipUnless(os.environ.get("test_with_gpu", False), "require GPU")
class TestProxy(unittest.TestCase):
    def test_send(self):
        ndev = 2
        uid = nccl.get_unique_id()
        with ProcessPoolExecutor(max_workers=2) as executor:
            st = executor.submit(sender, ndev, uid, 0)
            rt = executor.submit(receiver, ndev, uid, 1)

            tasks = [st, rt]

        results = []
        for t in as_completed(tasks):
            results.append(t.result())

        assert np.allclose(results[0], results[1])

    def test_broadcast(self):
        ndev = 3
        uid = nccl.get_unique_id()
        with ProcessPoolExecutor(max_workers=ndev) as executor:
            tasks = [executor.submit(broadcast, ndev, uid, dev_id) for dev_id in range(ndev)]

        results = []
        for t in as_completed(tasks):
            results.append(t.result())

        assert np.allclose(results[0], results[1])
        assert np.allclose(results[0], results[2])


if __name__ == "__main__":
    unittest.main()



# -*- coding:utf8 -*-

from copy import copy
import numpy as np
from typing import *

from QuICT.core import Circuit
from QuICT.core.gate import *


class TikzDrawer:
    wire_gap = 0.8
    level_gap = 1.0
    rectangle_size = level_gap / 2

    @classmethod
    def circuit_layers(cls, circuit: Circuit) -> Iterable[List[BasicGate]]:
        ret = []
        cur_level = []
        cur_level_occ = set([])

        for gate in circuit.gates:
            gate: BasicGate

            gate_min = np.min(gate.affectArgs)
            gate_max = np.max(gate.affectArgs)
            is_not_occupied = True
            for bit in range(gate_min, gate_max + 1):
                if bit in cur_level_occ:
                    is_not_occupied = False
                cur_level_occ.add(bit)

            if is_not_occupied:
                cur_level.append(gate)
            else:
                ret.append(copy(cur_level))
                cur_level = [gate]
                cur_level_occ = set(list(range(gate_min, gate_max + 1)))

        ret.append(cur_level)
        return ret

    @classmethod
    def layer_print(cls, circuit: Circuit):
        layers = cls.circuit_layers(circuit)
        for idx, layer in enumerate(layers):
            print("=" * 40)
            print(f"[Layer {idx}]")
            for gate in layer:
                gate.print_info()

    @classmethod
    def get_center_pos(cls, qubit: int, level: int) -> Tuple[float, float]:

        y = 0 - cls.wire_gap * (qubit - 1)
        x = level * cls.level_gap + cls.level_gap / 2
        return x, y

    @classmethod
    def draw_1_qubit_gate(cls, gate: BasicGate, level: int) -> str:
        x, y = cls.get_center_pos(gate.targ, level)
        rectangle_size = cls.rectangle_size

        if isinstance(gate, XGate):
            name = "X"
        elif isinstance(gate, YGate):
            name = "Y"
        elif isinstance(gate, ZGate):
            name = "Z"
        elif isinstance(gate, HGate):
            name = "H"
        elif isinstance(gate, RxGate):
            name = "Rx"
        elif isinstance(gate, RyGate):
            name = "Ry"
        elif isinstance(gate, RzGate):
            name = "Rz"
        elif isinstance(gate, IDGate):
            name = "Id"
        else:
            name = "*"

        gate_specification = f" node[pos=.5] {{{name}}}"
        tikz_command = "% begin of single qubit rectangle\n"
        tikz_command += f"\\draw " \
                        f"({x - rectangle_size / 2},{y + rectangle_size / 2})" \
                        " rectangle " \
                        f"({x + rectangle_size / 2},{y - rectangle_size / 2})" + \
                        gate_specification + \
                        ";\n"

        tikz_command += "% end of single qubit rectangle\n\n"

        tikz_command += "% begin of single qubit wire next to rectangle\n"

        tikz_command += f"\\draw " \
                        f"({x - cls.level_gap / 2},{y}) " \
                        f"-- " \
                        f"({x - cls.rectangle_size / 2},{y})" \
                        f";\n" \
                        f"\\draw " \
                        f"({x + cls.rectangle_size / 2},{y}) " \
                        f"-- " \
                        f"({x + cls.level_gap / 2},{y}) " \
                        f";\n"
        tikz_command += "% end of single qubit wire next to rectangle\n\n"
        return tikz_command

    @classmethod
    def draw_2_qubit_gate(cls, gate: BasicGate, level: int) -> str:
        if isinstance(gate, CXGate):
            t_name = "X"
        elif isinstance(gate, CYGate):
            t_name = "Y"
        elif isinstance(gate, CZGate):
            t_name = "Z"
        elif isinstance(gate, CRzGate):
            t_name = "Rz"
        elif isinstance(gate, CHGate):
            t_name = "H"
        else:
            t_name = "*"

        t_x, t_y = cls.get_center_pos(gate.targ, level)
        rectangle_size = cls.rectangle_size

        gate_specification = f" node[pos=.5] {{{t_name}}}"

        tikz_command = "% begin of 2 qubit gate target rectangle\n"

        tikz_command += f"\\draw " \
                        f"({t_x - rectangle_size / 2},{t_y + rectangle_size / 2})" \
                        " rectangle " \
                        f"({t_x + rectangle_size / 2},{t_y - rectangle_size / 2})" + \
                        gate_specification + \
                        ";\n"

        tikz_command += "% end of 2 qubit gate target rectangle\n\n"

        tikz_command += "% begin of 2 qubit gate wire next to target rectangle\n"

        tikz_command += f"\\draw " \
                        f"({t_x - cls.level_gap / 2},{t_y}) " \
                        f"-- " \
                        f"({t_x - cls.rectangle_size / 2},{t_y}) " \
                        f";\n" \
                        f"\\draw " \
                        f"({t_x + cls.rectangle_size / 2},{t_y}) " \
                        f"-- " \
                        f"({t_x + cls.level_gap / 2},{t_y}) " \
                        f";\n"

        tikz_command += "% end of 2 qubit gate wire next to target rectangle\n\n"

        tikz_command += "% begin of 2 qubit gate control bit wire\n"

        c_x, c_y = cls.get_center_pos(gate.carg, level)
        tikz_command += f"\\draw " \
                        f"({c_x - cls.level_gap / 2},{c_y}) " \
                        f"-- " \
                        f"({c_x + cls.level_gap / 2},{c_y}) " \
                        f";\n" \
                        f"\\filldraw " \
                        f"({c_x},{c_y}) circle ({cls.level_gap / 20}) " \
                        f";\n"

        tikz_command += "% end of 2 qubit gate control bit wire\n\n"

        tikz_command += "% begin of 2 qubit gate connect wire\n"

        if t_y > c_y:
            t_y_offset = t_y - cls.rectangle_size / 2
        else:
            t_y_offset = t_y + cls.rectangle_size / 2

        tikz_command += f"\\draw " \
                        f"({c_x},{c_y}) " \
                        f"-- " \
                        f"({t_x},{t_y_offset}) " \
                        f";\n"

        tikz_command += "% end of 2 qubit gate connect wire\n"

        return tikz_command

    @classmethod
    def draw_gate(cls, gate: BasicGate, level: int) -> str:
        ret = []
        rg = len(gate.affectArgs)
        if rg == 1:
            ret.append(cls.draw_1_qubit_gate(gate, level))
        elif rg == 2:
            ret.append(cls.draw_2_qubit_gate(gate, level))
        else:
            raise NotImplementedError("Can only draw 1 or 2 qubit gate!!!")

        return "".join(ret)

    @classmethod
    def draw_layer_line(cls, layer_bit: List[int], level: int) -> str:
        ret = []
        for i, b_type in enumerate(layer_bit):
            x, y = cls.get_center_pos(i, level)
            tikz_command = "% begin of non occupied qubit wire\n"
            if b_type == 1:
                tikz_command += ""
            elif b_type == 2:
                tikz_command += ""
            else:
                tikz_command += f"\\draw " \
                                f"({x - cls.level_gap / 2},{y}) " \
                                f"-- " \
                                f"({x + cls.level_gap / 2},{y}) " \
                                f";\n"
            tikz_command += "% end of non occupied qubit wire\n\n"
            ret.append(tikz_command)
        return "".join(ret)

    @classmethod
    def draw_layer(cls, layer: List[BasicGate], level: int, qubit_num: int) -> str:
        layer_bit = [0 for _ in range(qubit_num)]
        ret = []
        for gate in layer:
            k = len(gate.affectArgs)
            for b in gate.affectArgs:
                layer_bit[b] = k
            ret.append(cls.draw_gate(gate, level))
        line_command = cls.draw_layer_line(layer_bit, level)
        ret.append(line_command)
        return "".join(ret)

    @classmethod
    def draw_empty_head_n_tail_wire(cls, level_max: int, qubit_num: int) -> str:
        ret = []
        for bit in range(qubit_num):
            x, y = cls.get_center_pos(bit, 0)
            tikz_command = "% begin of leading emtpy wires\n"
            tikz_command += f"\\draw " \
                            f"({x - cls.level_gap * 1.5 / 2},{y}) " \
                            f"-- " \
                            f"({x - cls.level_gap / 2},{y}) " \
                            f";\n"
            tikz_command += "% end of leading emtpy wires\n\n"

            x, y = cls.get_center_pos(bit, level_max)
            tikz_command += "% begin of tailing empty wires\n"
            tikz_command += f"\\draw " \
                            f"({x + cls.level_gap / 2},{y}) " \
                            f"-- " \
                            f"({x + cls.level_gap * 1.5 / 2},{y}) " \
                            f";\n"
            tikz_command += "% end of tailing empty wires\n\n"

            ret.append(tikz_command)

        return "".join(ret)

    @classmethod
    def run(cls, circuit: Circuit):
        ans = ["\\begin{tikzpicture}\n"]
        circuit_width = circuit.circuit_width()

        layer_cnt = 0
        for idx, layer in enumerate(cls.circuit_layers(circuit)):
            cmds = cls.draw_layer(layer, idx, circuit_width)
            ans.append(cmds)
            layer_cnt += 1

        ans.append(cls.draw_empty_head_n_tail_wire(layer_cnt - 1, circuit_width))

        ans.append("\\end{tikzpicture}")

        tikz_command = "".join(ans)
        # print()
        # print(tikz_command)
        return tikz_command



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/3/30 6:24
# @Author  : Han Yu
# @File    : _localDrawer.py

import collections
import math
import re

import numpy as np
from matplotlib import patches
from matplotlib.figure import Figure
from matplotlib import pyplot as plt

from QuICT.core.gate import *

from .ibmq_style import DefaultStyle

LINE_WIDTH = 1.5
FOLD = 26
cir_len = 0
FONT_SIZE = 13
SUB_FONT_SIZE = 8

WID = 0.65
HIG = 0.65
DEFAULT_SCALE = 4.3
PORDER_GATE = 5
PORDER_LINE = 3
PORDER_REGLINE = 2
PORDER_GRAY = 3
PORDER_TEXT = 6
PORDER_SUBP = 4


def pi_check(p, eps=1e-6, ndigits=3):
    p = float(p)
    if abs(p) < 1e-14:
        return '0'
    tp = p / np.pi
    if abs(tp) >= 1:
        if abs(tp % 1) < eps:
            val = int(round(tp))
            if val == 1:
                str_out = '$\\pi$'
            elif val == -1:
                str_out = '$-$$\\pi$'
            else:
                str_out = '{}$\\pi$'.format(val)
            return str_out

    tp = np.pi / p
    if abs(abs(tp) - abs(round(tp))) < eps:
        val = int(round(tp))
        if val > 0:
            str_out = '$\\pi$/{}'.format(val)
        else:
            str_out = '$-$$\\pi$/{}'.format(abs(val))
        return str_out

    abs_p = abs(p)
    N, D = np.meshgrid(np.arange(1, 64), np.arange(1, 64))
    frac = np.where(np.abs(abs_p - N / D * np.pi) < 1e-8)
    if frac[0].shape[0]:
        numer = int(frac[1][0]) + 1
        denom = int(frac[0][0]) + 1
        if p < 0:
            numer *= -1

        if numer == 1 and denom == 1:
            str_out = '$\\pi$'
        elif numer == -1 and denom == 1:
            str_out = '$-$$\\pi$'
        elif numer == 1:
            str_out = '$\\pi$/{}'.format(denom)
        elif numer == -1:
            str_out = '$-$$\\pi$/{}'.format(denom)
        elif denom == 1:
            str_out = '{}/$\\pi$'.format(numer)
        else:
            str_out = '{}$\\pi$/{}'.format(numer, denom)

        return str_out

    str_out = '%.{}g'.format(ndigits) % p
    return str_out


class circuit_layer(object):
    def __init__(self):
        self.pic = set()
        self.occupy = set()
        self.gates = []

    def addGate(self, gate: BasicGate) -> bool:
        Q_set = set(gate.cargs) | set(gate.targs)
        for element in range(min(Q_set), max(Q_set) + 1):
            if element in self.pic:
                return False
        for element in range(min(Q_set), max(Q_set) + 1):
            self.pic.add(element)
        self.occupy |= Q_set
        self.gates.append(gate)
        return True

    def checkGate(self, gate: BasicGate) -> bool:
        Q_set = set(gate.cargs) | set(gate.targs)
        for element in range(min(Q_set), max(Q_set) + 1):
            if element in self.pic:
                return False
        return True


class Anchor(object):
    def __init__(self, pos):
        self.pos = pos
        self.position = []
        self.anchor = 0

    def set_index(self, index, layer_width):
        h_pos = index % FOLD + 1
        if h_pos + (layer_width - 1) > FOLD:
            index = index + FOLD - (h_pos - 1)
        for i in range(layer_width):
            if index + i not in self.position:
                self.position.append(index + i)
        self.position.sort()

    def get_index(self):
        if len(self.position) > 0:
            return self.position[-1] + 1
        return 0

    def coord(self, position, layer_width, offset_x):
        h_pos = position % FOLD + 1
        if h_pos + (layer_width - 1) > FOLD:
            position = position + FOLD - (h_pos - 1)
        x = position % FOLD + 1 + 0.5 * (layer_width - 1)
        y = self.pos - (position // FOLD) * (cir_len + 1)
        self.anchor = position
        return x + offset_x, y


class PhotoDrawer(object):

    def __init__(self):
        self.style = DefaultStyle()
        # plt.figure() will always trigger interactive mode eagerly. Use matplotlib.figure.Figure's full OO interface.
        # refer:
        # https://stackoverflow.com/questions/18207563/using-interactive-and-non-interactive-backends-within-one-program
        self.figure = Figure()
        self.figure.patch.set_facecolor(color='#ffffff')
        self.ax = self.figure.add_subplot(111)
        self.ax.set_xticks([])
        self.ax.set_yticks([])
        self.scale = 1
        self.lwidth15 = 1.5 * self.scale
        self.lwidth2 = 2.0 * self.scale
        self.char_list = {' ': (0.0958, 0.0583), '!': (0.1208, 0.0729), '"': (0.1396, 0.0875),
                          '#': (0.2521, 0.1562), '$': (0.1917, 0.1167), '%': (0.2854, 0.1771),
                          '&': (0.2333, 0.1458), "'": (0.0833, 0.0521), '(': (0.1167, 0.0729),
                          ')': (0.1167, 0.0729), '*': (0.15, 0.0938), '+': (0.25, 0.1562),
                          ',': (0.0958, 0.0583), '-': (0.1083, 0.0667), '.': (0.0958, 0.0604),
                          '/': (0.1021, 0.0625), '0': (0.1875, 0.1167), '1': (0.1896, 0.1167),
                          '2': (0.1917, 0.1188), '3': (0.1917, 0.1167), '4': (0.1917, 0.1188),
                          '5': (0.1917, 0.1167), '6': (0.1896, 0.1167), '7': (0.1917, 0.1188),
                          '8': (0.1896, 0.1188), '9': (0.1917, 0.1188), ':': (0.1021, 0.0604),
                          ';': (0.1021, 0.0604), '<': (0.25, 0.1542), '=': (0.25, 0.1562),
                          '>': (0.25, 0.1542), '?': (0.1583, 0.0979), '@': (0.2979, 0.1854),
                          'A': (0.2062, 0.1271), 'B': (0.2042, 0.1271), 'C': (0.2083, 0.1292),
                          'D': (0.2312, 0.1417), 'E': (0.1875, 0.1167), 'F': (0.1708, 0.1062),
                          'G': (0.2312, 0.1438), 'H': (0.225, 0.1396), 'I': (0.0875, 0.0542),
                          'J': (0.0875, 0.0542), 'K': (0.1958, 0.1208), 'L': (0.1667, 0.1042),
                          'M': (0.2583, 0.1604), 'N': (0.225, 0.1396), 'O': (0.2354, 0.1458),
                          'P': (0.1812, 0.1125), 'Q': (0.2354, 0.1458), 'R': (0.2083, 0.1292),
                          'S': (0.1896, 0.1188), 'T': (0.1854, 0.1125), 'U': (0.2208, 0.1354),
                          'V': (0.2062, 0.1271), 'W': (0.2958, 0.1833), 'X': (0.2062, 0.1271),
                          'Y': (0.1833, 0.1125), 'Z': (0.2042, 0.1271), '[': (0.1167, 0.075),
                          '\\': (0.1021, 0.0625), ']': (0.1167, 0.0729), '^': (0.2521, 0.1562),
                          '_': (0.1521, 0.0938), '`': (0.15, 0.0938), 'a': (0.1854, 0.1146),
                          'b': (0.1917, 0.1167), 'c': (0.1646, 0.1021), 'd': (0.1896, 0.1188),
                          'e': (0.1854, 0.1146), 'f': (0.1042, 0.0667), 'g': (0.1896, 0.1188),
                          'h': (0.1896, 0.1188), 'i': (0.0854, 0.0521), 'j': (0.0854, 0.0521),
                          'k': (0.1729, 0.1083), 'l': (0.0854, 0.0521), 'm': (0.2917, 0.1812),
                          'n': (0.1896, 0.1188), 'o': (0.1833, 0.1125), 'p': (0.1917, 0.1167),
                          'q': (0.1896, 0.1188), 'r': (0.125, 0.0771), 's': (0.1562, 0.0958),
                          't': (0.1167, 0.0729), 'u': (0.1896, 0.1188), 'v': (0.1771, 0.1104),
                          'w': (0.2458, 0.1521), 'x': (0.1771, 0.1104), 'y': (0.1771, 0.1104),
                          'z': (0.1562, 0.0979), '{': (0.1917, 0.1188), '|': (0.1, 0.0604),
                          '}': (0.1896, 0.1188)}
        pass

    @staticmethod
    def resolution_layers(circuit):
        layers = [circuit_layer()]
        for gate in circuit.gates:
            for i in range(len(layers) - 1, -2, -1):
                if i == -1 or not layers[i].checkGate(gate):
                    if i + 1 >= len(layers):
                        layers.append(circuit_layer())
                    layers[i + 1].addGate(gate)
                    break
        return layers

    @staticmethod
    def get_parameter_str(params):
        strings = []
        for p in params:
            strings.append(pi_check(p))
        return ', '.join(strings)

    def draw_gate(self, point, text=None, p_string=None, fc=None):
        xpos, ypos = point

        c_len = False

        if p_string:
            if '$\\pi$' in p_string:
                pi_count = p_string.count('pi')
                subtext_len = len(p_string) - (4 * pi_count)
                if subtext_len > 10:
                    c_len = True
            else:
                c_len = len(p_string) > 10

        if c_len:
            subtext_len = len(p_string)
            if '$\\pi$' in p_string:
                pi_count = p_string.count('pi')
                subtext_len = subtext_len - (4 * pi_count)

            boxes_wide = round(max(subtext_len, len(text)) / 10, 1) or 1
            wid = WID * 1.5 * boxes_wide
            if wid < WID:
                wid = WID
        else:
            wid = WID

        if fc:
            _fc = fc
        elif text and text in self.style.dispcol:
            _fc = self.style.dispcol[text]
        else:
            _fc = self.style.gc
        box = patches.Rectangle(
            xy=(xpos - 0.5 * wid, ypos - 0.5 * HIG), width=wid, height=HIG,
            fc=_fc, ec=self.style.edge_color, linewidth=1.5, zorder=PORDER_GATE)
        self.ax.add_patch(box)

        if text:
            font_size = self.style.fs
            sub_font_size = self.style.sfs
            if text in ['reset']:
                disp_color = self.style.not_gate_lc
                sub_color = self.style.not_gate_lc
                font_size = self.style.math_fs

            else:
                disp_color = self.style.gt
                sub_color = self.style.sc

            if text in self.style.dispcol:
                disp_text = "${}$".format(self.style.disptex[text])
            else:
                disp_text = text

            if p_string:
                self.ax.text(xpos, ypos + 0.15 * HIG, disp_text, ha='center',
                             va='center', fontsize=font_size,
                             color=disp_color, clip_on=True,
                             zorder=PORDER_TEXT)
                self.ax.text(xpos, ypos - 0.3 * HIG, p_string, ha='center',
                             va='center', fontsize=sub_font_size,
                             color=sub_color, clip_on=True,
                             zorder=PORDER_TEXT)
            else:
                self.ax.text(xpos, ypos, disp_text, ha='center', va='center',
                             fontsize=font_size,
                             color=disp_color,
                             clip_on=True,
                             zorder=PORDER_TEXT)

    def draw_depth(self, start, end):
        start_pos = start % FOLD + 0.14
        end_pos = end % FOLD + 0.14
        if end_pos == 0:
            end_pos = FOLD

        start_layer = start // FOLD
        end_layer = end // FOLD

        if start_layer == end_layer:
            box = patches.Rectangle(
                xy=(start_pos - 0.75 * WID, - (start_layer + 1) * (cir_len + 1) + 1.5),
                width=end_pos - start_pos - 1 + 1.5 * WID, height=cir_len,
                fc='#FF0000', ec='#FF0000', linewidth=1.5, zorder=PORDER_GATE, fill=False, linestyle='dashed')
            self.ax.add_patch(box)
        else:
            self.ax.plot([start_pos - 0.75 * WID, start_pos - 0.75 * WID],
                         [- (start_layer + 1) * (cir_len + 1) + 1.5,
                          - (start_layer + 1) * (cir_len + 1) + 1.5 + cir_len],
                         color='#FF0000',
                         linewidth=1.5,
                         linestyle='dashed',
                         zorder=PORDER_GATE)
            self.ax.plot([start_pos - 0.75 * WID, FOLD + 0.14],
                         [- (start_layer + 1) * (cir_len + 1) + 1.5, - (start_layer + 1) * (cir_len + 1) + 1.5],
                         color='#FF0000',
                         linewidth=1.5,
                         linestyle='dashed',
                         zorder=PORDER_GATE)
            self.ax.plot([start_pos - 0.75 * WID, FOLD + 0.14],
                         [- (start_layer + 1) * (cir_len + 1) + 1.5 + cir_len,
                          - (start_layer + 1) * (cir_len + 1) + 1.5 + cir_len],
                         color='#FF0000',
                         linewidth=1.5,
                         linestyle='dashed',
                         zorder=PORDER_GATE)

            self.ax.plot([end_pos - 1 + 0.75 * WID, end_pos - 1 + 0.75 * WID],
                         [- (end_layer + 1) * (cir_len + 1) + 1.5,
                          - (end_layer + 1) * (cir_len + 1) + 1.5 + cir_len],
                         color='#FF0000',
                         linewidth=1.5,
                         linestyle='dashed',
                         zorder=PORDER_GATE)
            self.ax.plot([-0.75 * WID, end_pos - 1 + 0.75 * WID],
                         [- (end_layer + 1) * (cir_len + 1) + 1.5,
                          - (end_layer + 1) * (cir_len + 1) + 1.5],
                         color='#FF0000',
                         linewidth=1.5,
                         linestyle='dashed',
                         zorder=PORDER_GATE)
            self.ax.plot([-0.75 * WID, end_pos - 1 + 0.75 * WID],
                         [- (end_layer + 1) * (cir_len + 1) + 1.5 + cir_len,
                          - (end_layer + 1) * (cir_len + 1) + 1.5 + cir_len],
                         color='#FF0000',
                         linewidth=1.5,
                         linestyle='dashed',
                         zorder=PORDER_GATE)
            for i in range(start_layer + 1, end_layer):
                self.ax.plot([-0.75 * WID, FOLD + 0.14],
                             [- (i + 1) * (cir_len + 1) + 1.5 + cir_len,
                              - (i + 1) * (cir_len + 1) + 1.5 + cir_len],
                             color='#FF0000',
                             linewidth=1.5,
                             linestyle='dashed',
                             zorder=PORDER_GATE)
                self.ax.plot([-0.75 * WID, FOLD + 0.14],
                             [- (i + 2) * (cir_len + 1) + 1.5 + 1 + cir_len,
                              - (i + 2) * (cir_len + 1) + 1.5 + 1 + cir_len],
                             color='#FF0000',
                             linewidth=1.5,
                             linestyle='dashed',
                             zorder=PORDER_GATE)

    def draw_line(self, xy0, xy1, lc=None, ls=None, zorder=PORDER_LINE):
        x0, y0 = xy0
        x1, y1 = xy1
        if lc is None:
            linecolor = self.style.lc
        else:
            linecolor = lc
        if ls is None:
            linestyle = 'solid'
        else:
            linestyle = ls

        if linestyle == 'doublet':
            theta = np.arctan2(np.abs(x1 - x0), np.abs(y1 - y0))
            dx = 0.05 * WID * np.cos(theta)
            dy = 0.05 * WID * np.sin(theta)
            self.ax.plot([x0 + dx, x1 + dx], [y0 + dy, y1 + dy],
                         color=linecolor,
                         linewidth=2,
                         linestyle='solid',
                         zorder=zorder)
            self.ax.plot([x0 - dx, x1 - dx], [y0 - dy, y1 - dy],
                         color=linecolor,
                         linewidth=2,
                         linestyle='solid',
                         zorder=zorder)
        else:
            self.ax.plot([x0, x1], [y0, y1],
                         color=linecolor,
                         linewidth=2,
                         linestyle=linestyle,
                         zorder=zorder)

    def draw_custom_gate(self, xy, cxy=None, fc=None, wide=True, text=None,
                         subtext=None):
        xpos = min([x[0] for x in xy])
        ypos = min([y[1] for y in xy])
        ypos_max = max([y[1] for y in xy])

        if cxy:
            ypos = min([y[1] for y in cxy])
        if wide:
            if subtext:
                boxes_length = round(max([len(text), len(subtext)]) / 6) or 1
            else:
                boxes_length = math.ceil(len(text) / 6) or 1
            wid = WID * 2.5 * boxes_length
        else:
            wid = WID

        if fc:
            _fc = fc
        else:
            if self.style.name != 'bw':
                if self.style.gc != DefaultStyle().gc:
                    _fc = self.style.gc
                else:
                    _fc = self.style.dispcol['multi']
            else:
                _fc = self.style.gc

        qubit_span = abs(ypos) - abs(ypos_max) + 1
        height = HIG + (qubit_span - 1)
        box = patches.Rectangle(
            xy=(xpos - 0.5 * wid, ypos - .5 * HIG),
            width=wid, height=height,
            fc=_fc,
            ec=self.style.dispcol['multi'],
            linewidth=1.5, zorder=PORDER_GATE)
        self.ax.add_patch(box)
        # Annotate inputs
        for bit, y in enumerate([x[1] for x in xy]):
            self.ax.text(xpos - 0.45 * wid, y, str(bit), ha='left', va='center',
                         fontsize=self.style.fs, color=self.style.gt,
                         clip_on=True, zorder=PORDER_TEXT)

        if text:
            disp_text = text
            if subtext:
                self.ax.text(xpos, ypos + 0.5 * height, disp_text, ha='center',
                             va='center', fontsize=self.style.fs,
                             color=self.style.gt, clip_on=True,
                             zorder=PORDER_TEXT)
                self.ax.text(xpos, ypos + 0.3 * height, subtext, ha='center',
                             va='center', fontsize=self.style.sfs,
                             color=self.style.sc, clip_on=True,
                             zorder=PORDER_TEXT)
            else:
                self.ax.text(xpos, ypos + .5 * (qubit_span - 1), disp_text,
                             ha='center',
                             va='center',
                             fontsize=self.style.fs,
                             color=self.style.gt,
                             clip_on=True,
                             zorder=PORDER_TEXT,
                             wrap=True)

    def draw_measure(self, xy):
        x, y = xy
        self.draw_gate(xy, fc=self.style.dispcol['meas'])

        arc = patches.Arc(xy=(x, y - 0.15 * HIG), width=WID * 0.7,
                          height=HIG * 0.7, theta1=0, theta2=180, fill=False,
                          ec=self.style.not_gate_lc, linewidth=2,
                          zorder=PORDER_GATE)
        self.ax.add_patch(arc)
        self.ax.plot([x, x + 0.35 * WID],
                     [y - 0.15 * HIG, y + 0.20 * HIG],
                     color=self.style.not_gate_lc, linewidth=2, zorder=PORDER_GATE)

    def draw_ctrl_qubit(self, xy, fc=None, ec=None):
        if self.style.gc != DefaultStyle().gc:
            fc = self.style.gc
            ec = self.style.gc
        if fc is None:
            fc = self.style.lc
        if ec is None:
            ec = self.style.lc
        xpos, ypos = xy
        box = patches.Circle(xy=(xpos, ypos), radius=WID * 0.15,
                             fc=fc, ec=ec,
                             linewidth=1.5, zorder=PORDER_GATE)
        self.ax.add_patch(box)

    def draw_tgt_qubit(self, xy, fc=None, ec=None, ac=None, add_width=None):
        if self.style.gc != DefaultStyle().gc:
            fc = self.style.gc
            ec = self.style.gc
        if fc is None:
            fc = self.style.dispcol['target']
        if ec is None:
            ec = self.style.lc
        if ac is None:
            ac = self.style.lc
        if add_width is None:
            add_width = 0.35

        linewidth = 2

        if self.style.dispcol['target'] == '#ffffff':
            add_width = self.style.colored_add_width

        xpos, ypos = xy

        box = patches.Circle(xy=(xpos, ypos), radius=HIG * 0.35,
                             fc=fc, ec=ec, linewidth=linewidth,
                             zorder=PORDER_GATE)
        self.ax.add_patch(box)
        # add '+' symbol
        self.ax.plot([xpos, xpos], [ypos - add_width * HIG,
                                    ypos + add_width * HIG],
                     color=ac, linewidth=linewidth, zorder=PORDER_GATE + 1)

        self.ax.plot([xpos - add_width * HIG, xpos + add_width * HIG],
                     [ypos, ypos], color=ac, linewidth=linewidth,
                     zorder=PORDER_GATE + 1)

    def _get_text_width(self, text, fontsize, param=False):
        if not text:
            return 0.0

        math_mode_match = re.compile(r"(?<!\\)\$(.*)(?<!\\)\$").search(text)
        num_underscores = 0
        num_carets = 0
        if math_mode_match:
            math_mode_text = math_mode_match.group(1)
            num_underscores = math_mode_text.count('_')
            num_carets = math_mode_text.count('^')

        # If there are subscripts or superscripts in mathtext string
        # we need to account for that spacing by manually removing
        # from text string for text length
        if num_underscores:
            text = text.replace('_', '', num_underscores)
        if num_carets:
            text = text.replace('^', '', num_carets)

        # This changes hyphen to + to match width of math mode minus sign.
        if param:
            text = text.replace('-', '+')

        f = 0 if fontsize == self.style else 1
        sum_text = 0.0
        for c in text:
            try:
                if c not in ["$", "\\"]:
                    sum_text += self.char_list[c][f]
            except KeyError:
                # if non-ASCII char, use width of 'c', an average size
                sum_text += self.char_list['c'][f]

        return sum_text

    def draw_multiqubit_gate(self, xy, fc=None, ec=None, gt=None, sc=None, text='', subtext=''):
        xpos = min([x[0] for x in xy])
        ypos = min([y[1] for y in xy])
        ypos_max = max([y[1] for y in xy])
        fs = self.style.fs
        sfs = self.style.sfs

        # added .21 is for qubit numbers on the left side
        text_width = self._get_text_width(text, fs) + .21 * 2
        sub_width = self._get_text_width(subtext, sfs, param=True) + .21 * 2
        wid = max((text_width, sub_width, WID))

        qubit_span = abs(ypos) - abs(ypos_max) + 1
        height = HIG + (qubit_span - 1)
        box = patches.Rectangle(
            xy=(xpos - 0.5 * wid, ypos - 0.5 * HIG), width=wid, height=height,
            fc=fc, ec=ec, linewidth=self.lwidth15, zorder=PORDER_GATE)
        self.ax.add_patch(box)

        # annotate inputs
        for bit, y in enumerate([x[1] for x in xy]):
            self.ax.text(xpos + .07 - 0.5 * wid, y, str(bit), ha='left', va='center',
                         fontsize=fs, color=gt,
                         clip_on=True, zorder=PORDER_TEXT)
        if text:
            if subtext:
                self.ax.text(xpos + .11, ypos + 0.4 * height, text, ha='center',
                             va='center', fontsize=fs,
                             color=gt, clip_on=True,
                             zorder=PORDER_TEXT)
                self.ax.text(xpos + .11, ypos + 0.2 * height, subtext, ha='center',
                             va='center', fontsize=sfs,
                             color=sc, clip_on=True,
                             zorder=PORDER_TEXT)
            else:
                self.ax.text(xpos + .11, ypos + .5 * (qubit_span - 1), text,
                             ha='center', va='center', fontsize=fs,
                             color=gt, clip_on=True,
                             zorder=PORDER_TEXT, wrap=True)

    def draw_swap(self, xy):
        xpos, ypos = xy
        color = self.style.dispcol['swap']
        self.ax.plot([xpos - 0.20 * WID, xpos + 0.20 * WID],
                     [ypos - 0.20 * WID, ypos + 0.20 * WID],
                     color=color, linewidth=2, zorder=PORDER_LINE + 1)
        self.ax.plot([xpos - 0.20 * WID, xpos + 0.20 * WID],
                     [ypos + 0.20 * WID, ypos - 0.20 * WID],
                     color=color, linewidth=2, zorder=PORDER_LINE + 1)

    def draw_linefeed_mark(self, xy):
        xpos, ypos = xy

        self.ax.plot([xpos - .1, xpos - .1],
                     [ypos, ypos - cir_len + 1],
                     color=self.style.lc, zorder=PORDER_LINE)
        self.ax.plot([xpos + .1, xpos + .1],
                     [ypos, ypos - cir_len + 1],
                     color=self.style.lc, zorder=PORDER_LINE)

    def draw_regs_sub(self, n_fold, offset_x=None, now=None, name_dict=None):
        for qreg in name_dict.values():
            if n_fold == 0:
                label = qreg['text']
            else:
                label = qreg['text']
            y = qreg['y'] - n_fold * (cir_len + 1)
            self.ax.text(offset_x, y, label, ha='right', va='center',
                         fontsize=1.25 * self.style.fs,
                         color=self.style.tc,
                         clip_on=True,
                         zorder=PORDER_TEXT)
            self.draw_line([offset_x + 0.5, y], [now['max_x'], y], zorder=PORDER_REGLINE)

    def run(self, circuit, filename=None, show_depth=False, save_file=False):
        global cir_len
        cir_len = circuit.width()
        name_dict = collections.OrderedDict()
        now = {
            'max_x': 0,
            'max_y': 0,
        }
        anchors = {}
        max_name = 0
        for i in range(cir_len):
            name = f'$q_{{{i}}}$'
            max_name = max(max_name, len(name))
            name_dict[i] = {
                'y': -i,
                'text': name,
                'index': i
            }
            anchors[i] = Anchor(-i)
        offset_x = 0.18 * (max_name - 7) - 0.5

        layers = self.resolution_layers(circuit)
        layer_position = []
        position = 0
        for layer in layers:
            layer_width = 1

            for gate in layer.gates:
                if gate.type == GateType.perm or gate.type == GateType.unitary:
                    continue
                elif gate.params > 1:
                    param = self.get_parameter_str(gate.pargs)
                    if '$\\pi$' in param:
                        pi_count = param.count('pi')
                        len_param = len(param) - (4 * pi_count)
                    else:
                        len_param = len(param)
                    if len_param > len(gate.qasm_name):
                        box_width = math.floor(len(param) / 10)
                        if box_width <= 1:
                            box_width = 1

                        if layer_width <= box_width:
                            layer_width = box_width + 1

            for gate in layer.gates:
                coord = []
                for index in gate.cargs:
                    anchors[index].set_index(position, layer_width)
                    coord.append(anchors[index].coord(position, layer_width, offset_x))
                for index in gate.targs:
                    anchors[index].set_index(position, layer_width)
                    coord.append(anchors[index].coord(position, layer_width, offset_x))

                bottom = min(coord, key=lambda c: c[1])
                top = max(coord, key=lambda c: c[1])

                position = anchors[gate.targ].anchor

                param = None
                if gate.params > 0:
                    param = self.get_parameter_str(gate.pargs)

                if gate.type == GateType.perm:
                    name = gate.type.value
                    for coor in coord:
                        self.draw_gate(coor, name, '')
                    self.draw_line(bottom, top, lc=self.style.dispcol[name])
                elif gate.type == GateType.measure:
                    self.draw_measure(coord[0])
                elif gate.type == GateType.barrier:
                    self.draw_gate(coord[0], gate.qasm_name)
                elif len(coord) == 1:
                    name = gate.qasm_name
                    if param is not None:
                        p_string = '({})'.format(param)
                        self.draw_gate(coord[0], name, p_string)
                    else:
                        self.draw_gate(coord[0], name, '')
                elif len(coord) == 2:
                    if gate.type == GateType.cx:
                        if self.style.dispcol['cx'] != '#ffffff':
                            add_width = self.style.colored_add_width
                        else:
                            add_width = None
                        self.draw_ctrl_qubit(coord[0], fc=self.style.dispcol['cx'], ec=self.style.dispcol['cx'])

                        self.draw_tgt_qubit(coord[1], fc=self.style.dispcol['cx'], ec=self.style.dispcol['cx'],
                                            ac=self.style.dispcol['target'], add_width=add_width)
                        self.draw_line(coord[0], coord[1], lc=self.style.dispcol['cx'])
                    elif gate.type == GateType.swap:
                        self.draw_swap(coord[0])
                        self.draw_swap(coord[1])
                        self.draw_line(bottom, top, lc=self.style.dispcol['swap'])
                    elif gate.controls == 1:
                        disp = gate.qasm_name.replace('c', '')

                        color = None
                        if self.style.name != 'bw':
                            color = self.style.dispcol['multi']

                        self.draw_ctrl_qubit(coord[0], fc=color, ec=color)
                        if param:
                            self.draw_gate(coord[1],
                                           text=disp,
                                           fc=color,
                                           p_string='{}'.format(param))
                        else:
                            self.draw_gate(coord[1], text=disp,
                                           fc=color)
                        # add qubit-qubit wiring
                        self.draw_line(bottom, top, lc=color)
                    else:
                        if param:
                            subtext = '{}'.format(param)
                        else:
                            subtext = ''
                        self.draw_multiqubit_gate(coord, fc=self.style.dispcol['multi'], ec=self.style.dispcol['multi'],
                                                  gt=self.style.gt, sc=self.style.sc,
                                                  text=gate.qasm_name, subtext=subtext)
                elif len(coord) == 3:
                    if gate.type == GateType.ccx:
                        self.draw_ctrl_qubit(coord[0], fc=self.style.dispcol['multi'],
                                             ec=self.style.dispcol['multi'])
                        self.draw_ctrl_qubit(coord[1], fc=self.style.dispcol['multi'],
                                             ec=self.style.dispcol['multi'])
                        if self.style.name != 'bw':
                            self.draw_tgt_qubit(coord[2], fc=self.style.dispcol['multi'],
                                                ec=self.style.dispcol['multi'],
                                                ac=self.style.dispcol['target'])
                        else:
                            self.draw_tgt_qubit(coord[2], fc=self.style.dispcol['target'],
                                                ec=self.style.dispcol['multi'],
                                                ac=self.style.dispcol['multi'])
                        # add qubit-qubit wiring
                        self.draw_line(bottom, top, lc=self.style.dispcol['multi'])
                    elif gate.type == GateType.cswap:
                        self.draw_ctrl_qubit(coord[0], fc=self.style.dispcol['multi'],
                                             ec=self.style.dispcol['multi'])
                        self.draw_swap(coord[1])
                        self.draw_swap(coord[2])
                        self.draw_line(bottom, top, lc=self.style.dispcol['swap'])
                    elif gate.controls > 0:
                        for i in range(gate.controls):
                            self.draw_ctrl_qubit(coord[i], fc=self.style.dispcol['multi'],
                                                 ec=self.style.dispcol['multi'])
                        if param:
                            subtext = '{}'.format(param)
                        else:
                            subtext = ''
                        if gate.targets >= 2:
                            self.draw_multiqubit_gate(coord[gate.controls:], fc=self.style.dispcol['multi'],
                                                      ec=self.style.dispcol['multi'],
                                                      gt=self.style.gt, sc=self.style.sc,
                                                      text=gate.qasm_name, subtext=subtext)
                        else:
                            self.draw_gate(coord[-1],
                                           text=gate.qasm_name[gate.controls:],
                                           fc=self.style.dispcol['multi'],
                                           p_string=subtext)
                        self.draw_line(bottom, top, lc=self.style.dispcol['swap'])
                    else:
                        if param:
                            subtext = '{}'.format(param)
                        else:
                            subtext = ''
                        self.draw_multiqubit_gate(coord, fc=self.style.dispcol['multi'],
                                                  ec=self.style.dispcol['multi'],
                                                  gt=self.style.gt, sc=self.style.sc,
                                                  text=gate.qasm_name, subtext=subtext)
                elif len(coord) > 3:
                    self.draw_multiqubit_gate(
                        coord, fc=self.style.dispcol['multi'],
                        ec=self.style.dispcol['multi'],
                        gt=self.style.gt, sc=self.style.sc,
                        text=gate.qasm_name, subtext=""
                    )

            layer_position.append(position)
            position = position + layer_width
        layer_position.append(position)

        temp_ac = [anchors[i].get_index() for i in range(cir_len)]
        max_anc = max(temp_ac)
        n_fold = max(0, max_anc - 1) // FOLD
        # window size
        if max_anc > FOLD > 0:
            now['max_x'] = FOLD + 1 + offset_x
            now['max_y'] = (n_fold + 1) * (cir_len + 1) - 1
        else:
            now['max_x'] = max_anc + 1 + offset_x
            now['max_y'] = cir_len
        # add horizontal lines

        now['max_x'] = max(5, now['max_x'])

        for ii in range(n_fold + 1):
            self.draw_regs_sub(ii, offset_x, now, name_dict)
        # draw gate number
        if self.style.index:
            for ii in range(max_anc):
                if FOLD > 0:
                    x_coord = ii % FOLD + 1
                    y_coord = - (ii // FOLD) * (cir_len + 1) + 0.7
                else:
                    x_coord = ii + 1
                    y_coord = 0.7
                self.ax.text(x_coord, y_coord, str(ii + 1), ha='center',
                             va='center', fontsize=self.style.sfs,
                             color=self.style.tc, clip_on=True,
                             zorder=PORDER_TEXT)
        if show_depth and len(layers) >= 2:
            depth_origin = []
            last_i = 0
            last_occupy = layers[0].occupy
            for i in range(1, len(layers)):
                if len(layers[i].occupy & last_occupy) == 0:
                    last_occupy |= layers[i].occupy
                else:
                    if i - last_i >= 2:
                        depth_origin.append((last_i, i - 1))
                    last_i = i
                    last_occupy = layers[i].occupy
            for depth in depth_origin:
                start, end = depth
                self.draw_depth(layer_position[start], layer_position[end + 1])

        _xl = - self.style.margin[0]
        _xr = now['max_x'] + self.style.margin[1]
        _yb = - now['max_y'] - self.style.margin[2] + 1 - 0.5
        _yt = self.style.margin[3] + 0.5
        self.ax.set_xlim(_xl, _xr)
        self.ax.set_ylim(_yb, _yt)
        fig_w = _xr - _xl
        fig_h = _yt - _yb
        if self.style.figwidth < 0.0:
            self.style.figwidth = fig_w * 4.3 * self.style.fs / 72 / WID
        self.figure.set_size_inches(self.style.figwidth, self.style.figwidth * fig_h / fig_w)

        if save_file:
            filename = f"{circuit.name}.jpg" if filename is None else filename

        if filename is not None:
            self.figure.savefig(filename, dpi=self.style.dpi,
                                bbox_inches='tight')



from QuICT.tools.exception import QuICTException


class CircuitAppendError(QuICTException):
    """ Circuit Append Operator Error. """
    def __init__(self, msg: str = None):
        super().__init__(1004, msg)


class CircuitReplaceError(QuICTException):
    """ Circuit Replace Gate Error. """
    def __init__(self, msg: str = None):
        super().__init__(1005, msg)


class CircuitSpecialAppendError(QuICTException):
    """ Circuit special append including Random Append and Supremacy Append. """
    def __init__(self, msg: str = None):
        super().__init__(1006, msg)


class CircuitDAGError(QuICTException):
    """ DAG Circuit Error. """
    def __init__(self, msg: str = None):
        super().__init__(1007, msg)


class CircuitDrawError(QuICTException):
    """ Circuit Draw Error. """
    def __init__(self, msg: str = None):
        super().__init__(1008, msg)


class QASMError(QuICTException):
    """ QASM Error. """
    def __init__(self, other: str = None, line: str = None, file: str = None):
        msg = "Qasm error:{} \n in line:{} \n error file:{}".format(other, line, file)
        super().__init__(1009, msg)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/9/11 18:20
# @Author  : Xiaoquan Xu
# @File    : rhf_classical_simulation.py

import os
import numpy as np
import scipy.linalg as splin

from QuICT.chemistry.simulation.parametertensor import *
from QuICT.chemistry.simulation.moleculardata import MolecularData
from QuICT.chemistry.simulation.rhf_objective import RHFObjective

DATA_DIR = os.path.dirname(os.path.abspath(__file__)) + '/molecular_data'


def rhf_classical_simulation(n_atoms, distance, data_dir=None):
    if data_dir is None:
        data_dir = DATA_DIR
    data_dir += "/hydrogen_chains"
    data_dir += "/h_" + str(n_atoms) + "_sto-3g"
    data_dir += "/bond_distance_" + str(distance)
    molfile = data_dir + "/H" + str(n_atoms)
    molfile += "_sto-3g_singlet_linear_r-" + str(distance)
    moldata = MolecularData(molfile)

    S = np.load(data_dir + "/overlap.npy")
    Hcore = np.load(data_dir + "/h_core.npy")
    TEI = np.einsum("psqr", np.load(data_dir + "/tei.npy"))  # (1, 1, 0, 0)

    _, X = splin.eigh(Hcore, S)
    obi = obi_basis_rotation(Hcore, X)
    tbi = tbi_basis_rotation(TEI, X)
    # print(tbi)
    molecular_hamiltonian = generate_hamiltonian(moldata.nuclear_repulsion, obi, tbi)
    # print(molecular_hamiltonian.const)
    # print(molecular_hamiltonian.obi)
    # print(molecular_hamiltonian.tbi)
    # print(moldata.n_electrons)

    rhf_objective = RHFObjective(molecular_hamiltonian, moldata.n_electrons)
    result = rhf_objective.minimization()

    return rhf_objective, moldata, result, obi, tbi


if __name__ == "__main__":
    rhf_objective, moldata, result, obi, tbi = rhf_classical_simulation(6, 1.3)
    assert moldata.n_orbitals == 6
    assert moldata.n_electrons == 6
    assert moldata.one_body_integrals is None
    assert moldata.two_body_integrals is None

    assert result['fun'] == -2.924060484972225
    assert result['success'] is True
    assert result['nit'] == 7

    rhf_objective, moldata, result, obi, tbi = rhf_classical_simulation(3, 2.5)
    # actually H_3 is H_3_plus
    assert result['fun'] == -0.8062049864970052
    assert result['success'] is True
    assert result['nit'] == 2



import numpy as np


def schmidt_decompose(state_vector: np.ndarray, A_qubits: int):
    """
    A quantum state |psi> of a composite system A, B could be decomposed as |psi> = sum lambda_i |i_A> |i_B>,
    where lambda_i are non-negative real numbers, sum lambda_i^2 = 1, and |i_A>, |i_B> are orthonormal states.
    Such decomposition is called Schmidt decomposition, the lambda_i are called Schmidt coefficients,
    the number of non-zero lambda_i is called Schmidt number, and |i_A>, |i_B> are called Schmidt bases.

    In this function, we restrict A and B to be the first several qubits and the last several qubits respectively.

    Args:
        state_vector(np.ndarray): the state vector of the given state
        A_qubits(int): the number of the first qubits corresponding to A

    Returns:
        np.ndarray, np.ndarray, np.ndarray: lambda_i, |i_A>, |i_B> respectively
    """
    state_vector = np.array(state_vector)
    num_qubits = int(np.log2(state_vector.size))
    assert state_vector.ndim == 1 and 1 << num_qubits == state_vector.size,\
        ValueError('Quantum state should be an array with length 2^n')
    assert isinstance(A_qubits, int) and 0 < A_qubits and A_qubits < num_qubits,\
        ValueError('System A should have less qubits than total')
    B_qubits = num_qubits - A_qubits

    state_vector = state_vector.reshape(1 << A_qubits, 1 << B_qubits)
    U, d, V = np.linalg.svd(state_vector)
    return d, U.T[:len(d)], V[:len(d)]



from itertools import product
import numpy as np
from typing import List, Tuple

from .utils import is_kraus_ops, NoiseChannel
from QuICT.core.gate import ID, X, Y, Z
from QuICT.ops.linalg.cpu_calculator import tensor, dot
from QuICT.tools.exception.core import (
    TypeError, ValueError, KrausError, NoiseApplyError,
    PauliNoiseUnmatchedError, DamplingNoiseMixedProbExceedError
)


class QuantumNoiseError:
    """ The based class for quantum noise error.

    Example:
        quantum_error = QuantumNoiseError([(error_matrix1, 0.4), (error_matrix2, 0.6)])

    Args:
        ops (List[Tuple(error_matrix: np.ndarray, prob: float)]): The noise error operators, the sum of
            the probabilities should be 1.
    """
    @property
    def kraus(self):
        """ The kraus operator """
        return self._kraus_operators

    @property
    def kraus_ct(self):
        """ The kraus operator's conjugate transpose. """
        return self._kraus_conj_ops

    @property
    def operators(self):
        """ The noise error's matrix. """
        return self._operators

    @operators.setter
    def operators(self, operators: List[Tuple[np.ndarray, float]]):
        self._operators = operators
        self._kraus_operators = [np.sqrt(prob) * mat for mat, prob in self._operators]
        assert is_kraus_ops(self._kraus_operators), KrausError("There is not a Kraus operator.")
        self._kraus_conj_ops = [np.transpose(k).conjugate() for k in self._kraus_operators]

    @property
    def precision(self):
        """ The precision of noise error. """
        return self._precision

    @property
    def type(self):
        """ The type of noise error. """
        return self._type

    @type.setter
    def type(self, channel: NoiseChannel):
        assert isinstance(channel, NoiseChannel), TypeError("QuantumNoiseError.type", "NoiseChannel", type(channel))
        self._type = channel

    @property
    def qubits(self):
        """ The number of qubits. """
        return self._qubits

    def __init__(self, ops: List[Tuple[np.ndarray, float]]):
        assert isinstance(ops, list), TypeError("QuantumNoiseError.ops", "list", type(ops))

        # Initial error matrix and probability
        self._operators = ops
        self._kraus_operators = [np.sqrt(prob) * mat for mat, prob in self._operators]
        self._kraus_conj_ops = [np.transpose(k).conjugate() for k in self._kraus_operators]
        if not is_kraus_ops(self._kraus_operators):
            raise KrausError("There is not a Kraus operator.")

        self._qubits = int(np.log2(self._kraus_operators[0].shape[0]))
        self._precision = ops[0][0].dtype
        self._type = NoiseChannel.unitary

    def __str__(self):
        ne_str = f"{self.type.value} with {self.qubits} qubits.\n"
        for prob, nm in enumerate(self._operators):
            ne_str += f"Noise with probability {prob}: {nm}\n"

        return ne_str

    def expand(self, extend_qubits: int, change_itself: bool = False):
        """ Expand noise with size of extend_qubits. """
        extra_identity = np.identity(2 ** extend_qubits, dtype=self._precision)
        expand_operator = []
        for mat, prob in self._operators:
            expand_matrix = tensor(mat, extra_identity)
            expand_operator.append((expand_matrix, prob))

        if change_itself:
            self.operators = expand_matrix
        else:
            return QuantumNoiseError(expand_operator)

    def compose(self, other):
        """ generate composed noise error with self and other. """
        assert isinstance(other, QuantumNoiseError), TypeError(
            "QuantumNoiseError.compose", "QuantumNoiseError", type(other)
        )

        left_operator, right_operator = self.operators, other.operators
        if self.qubits > other.qubits:
            right_operator = other.expand(self.qubits - other.qubits).operators
        elif self.qubits < other.qubits:
            left_operator = self.expand(other.qubits - self.qubits).operators

        composed_ops = []
        for noise_matrix, prob in left_operator:
            for other_noise_matrix, other_prob in right_operator:
                composed_ops.append((dot(noise_matrix, other_noise_matrix), prob * other_prob))

        return QuantumNoiseError(composed_ops)

    def tensor(self, other):
        """ generate tensor noise error with self and other. """
        assert isinstance(other, QuantumNoiseError), TypeError(
            "QuantumNoiseError.compose", "QuantumNoiseError", type(other)
        )

        composed_ops = []
        for noise_matrix, prob in self._operators:
            for other_noise_matrix, other_prob in other.operators:
                composed_ops.append((tensor(noise_matrix, other_noise_matrix), prob * other_prob))

        return QuantumNoiseError(composed_ops)

    def apply_to_gate(self, matrix: np.ndarray):
        """ generate kraus operator with given gate's matrix. """
        assert matrix.shape == (2 ** self._qubits, 2 ** self._qubits), NoiseApplyError(
            f"The shape of given gate's matrix should equal to current noise's qubit number {self._qubits}."
        )

        return [dot(ops, matrix) for ops in self.kraus]

    def prob_mapping_operator(self, prob: float) -> np.ndarray:
        """ Return the related noise error's matrix with given probability. """
        for matrix, error_prob in self._operators:
            if prob < error_prob:
                return matrix

            prob -= error_prob


class PauliError(QuantumNoiseError):
    """ Pauli Error; Including Bit Flip and Phase Flip

    Example:
        PauliError([('i', 0.3), ('x', 0.7)])

    Args:
        ops (List[Tuple[str, float]]): The operators for pauli error.
        num_qubits (int): The number of target qubits. Defaults to 1.
    """
    _BASED_GATE_MATRIX = {
        'i': ID.matrix,
        'x': X.matrix,
        'y': Y.matrix,
        'z': Z.matrix
    }

    def __init__(self, ops: List[Tuple[str, float]], num_qubits: int = 1):
        if not isinstance(ops, (list, tuple)):
            raise TypeError("PauliError.ops", "[list, tuple]", type(ops))

        # Building noise error operators List[(pauli_matrix, prob)].
        operators = []
        for op, prob in ops:
            if prob < 0 or prob > 1:
                raise ValueError("PauliError.ops.prob", "[0, 1]", prob)

            if len(op) != num_qubits and num_qubits > 1:
                raise PauliNoiseUnmatchedError("Pauli Error get wrong input.")

            based_matrix = None
            for g in op:
                if g not in list(self._BASED_GATE_MATRIX.keys()):
                    raise ValueError("PauliError.ops", "[i, x, y, z]", g)

                if based_matrix is None:
                    based_matrix = self._BASED_GATE_MATRIX[g]
                    continue

                if num_qubits > 1:
                    based_matrix = tensor(based_matrix, self._BASED_GATE_MATRIX[g])
                else:
                    based_matrix = dot(based_matrix, self._BASED_GATE_MATRIX[g])

            operators.append((based_matrix, prob))

        # Initial PauilError
        super().__init__(operators)
        self.type = NoiseChannel.pauil


class BitflipError(PauliError):
    """ Special Case for PauilError, with fixed Pauil Operator:
        [('x', prob), ('i', 1 - prob)]

    Args:
        prob (float): The probability to flip the qubit.
    """
    def __init__(self, prob: float):
        ops = [('i', 1 - prob), ('x', prob)]

        super().__init__(ops)


class PhaseflipError(PauliError):
    """ Special Case for PauilError, with fixed Pauil Operator:
        [('z', prob), ('i', 1 - prob)]

    Args:
        prob (float): The probability to flip the phase.
    """
    def __init__(self, prob: float):
        ops = [('i', 1 - prob), ('z', prob)]

        super().__init__(ops)


class PhaseBitflipError(PauliError):
    """ Special Case for PauilError, with fixed Pauil Operator:
        [('y', prob), ('i', 1 - prob)]

    Args:
        prob (float): The probability to flip the qubit and phase.
    """
    def __init__(self, prob: float):
        ops = [('i', 1 - prob), ('y', prob)]

        super().__init__(ops)


class DepolarizingError(PauliError):
    """ The Depolarizing Error

    Args:
        prob (float): The probability of depolarizing.
        num_qubits (int, optional): The number of qubits have depolarizing error. Defaults to 1.
    """
    def __init__(self, prob: float, num_qubits: int = 1):
        if not isinstance(prob, float):
            raise TypeError("DepolarizingError.prob", "float", type(prob))

        if not isinstance(num_qubits, int):
            raise TypeError("DepolarizingError.number_qubits", "int", type(prob))

        assert num_qubits >= 1, ValueError("DepolarizingError.num_qubits", ">= 1", num_qubits)
        num_ops = 4 ** num_qubits
        max_prob = num_ops / (num_ops - 1)
        if prob < 0 or prob > max_prob:
            raise ValueError("DepolarizingError.prob", f"[0, {max_prob}]", prob)

        probs = [1 - prob / max_prob] + [prob / num_ops] * (num_ops - 1)
        ops = [''.join(i) for i in product(self._BASED_GATE_MATRIX.keys(), repeat=num_qubits)]

        super().__init__(list(zip(ops, probs)), num_qubits)
        self.type = NoiseChannel.depolarizing


class DampingError(QuantumNoiseError):
    """ Amplitude Damping, Phase Damping and Amp-Phase Damping

        a = amp, b = phase, p = state prob \n
        A0 = sqrt(1 - p1) * [[1, 0], [0, sqrt(1 - a - b)]] \n
        A1 = sqrt(1 - p1) * [[0, sqrt(a)], [0, 0]] \n
        A2 = sqrt(1 - p1) * [[0, 0], [0, sqrt(b)]] \n
        B0 = sqrt(p1) * [[sqrt(1 - a - b), 0], [0, 1]] \n
        B1 = sqrt(p1) * [[0, 0], [sqrt(a), 0]] \n
        B2 = sqrt(p1) * [[sqrt(b), 0], [0, 0]] \n

    Args:
        amplitude_prob (float): The probability for damping amplitude.
        phase_prob (float): The probability for damping phase.
        dissipation_state (float, optional): The dissipation states. Defaults to 0.0.
    """
    def __init__(self, amplitude_prob: float, phase_prob: float, dissipation_state: float = 0.0):
        assert amplitude_prob >= 0 and amplitude_prob <= 1, ValueError(
            "DampingError.amplitude_prob", "[0, 1]", amplitude_prob
        )
        assert phase_prob >= 0 and phase_prob <= 1, ValueError("DampingError.phase_prob", "[0, 1]", phase_prob)
        assert dissipation_state >= 0 and dissipation_state <= 1, ValueError(
            "DampingError.dissipation_state", "[0, 1]", dissipation_state
        )

        if amplitude_prob + phase_prob > 1:
            raise DamplingNoiseMixedProbExceedError("the sum of amplitude and phase damping prob <= 1.")

        self.amplitude_prob = amplitude_prob
        self.phase_prob = phase_prob
        self.dissipation_state = dissipation_state

        operators = self._create_kraus_ops()
        super().__init__(operators)
        self.type = NoiseChannel.damping

    def __str__(self):
        damp_noise = f"{self.type.value} with 1 qubit\n"
        if self.amplitude_prob != 0:
            damp_noise += f"amplitude damping prob: {self.amplitude_prob} "

        if self.phase_prob != 0:
            damp_noise += f"phase damping prob: {self.phase_prob} "

        if self.dissipation_state != 0:
            damp_noise += f"prob of stable state: {self.dissipation_state}"

        return damp_noise

    def _create_kraus_ops(self):
        """ Create damping Kraus operators. """
        m0, m1 = np.sqrt(1 - self.dissipation_state), np.sqrt(self.dissipation_state)
        sqrt_amp, sqrt_phase = np.sqrt(self.amplitude_prob), np.sqrt(self.phase_prob)
        sqrt_ap = np.sqrt(1 - self.amplitude_prob - self.phase_prob)

        k0 = m0 * np.array([[1, 0], [0, sqrt_ap]], dtype=np.complex128)
        k1 = m0 * np.array([[0, sqrt_amp], [0, 0]], dtype=np.complex128)
        k2 = m0 * np.array([[0, 0], [0, sqrt_phase]], dtype=np.complex128)
        k3 = m1 * np.array([[sqrt_ap, 0], [0, 1]], dtype=np.complex128)
        k4 = m1 * np.array([[0, 0], [sqrt_amp, 0]], dtype=np.complex128)
        k5 = m1 * np.array([[sqrt_phase, 0], [0, 0]], dtype=np.complex128)
        kraus = [k0, k1, k2, k3, k4, k5]

        probs = [a * b for a, b in product(
            [1 - self.dissipation_state, self.dissipation_state],
            [1 - self.amplitude_prob - self.phase_prob, self.amplitude_prob, self.phase_prob]
        )]
        operator = []
        for idx, prob in enumerate(probs):
            if prob != 0:
                operator.append((kraus[idx], 1))

        return operator



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/4/27 下午5:50
# @Author  : Kaiqi Li
# @File    : proxy_simulator
import random
import cupy as cp
import numpy as np

from QuICT.simulation.state_vector import CircuitSimulator, ConstantStateVectorSimulator

from QuICT.core import Circuit
from QuICT.core.gate import BasicGate, GateType
from QuICT.core.operator import *
from QuICT.utility import Proxy
from QuICT.simulation.multi_nodes.data_switch import DataSwitcher
from QuICT.ops.gate_kernel.gate_function import mn_measureprob_calculator


class MultiNodesSimulator:
    """
    The simulator which using multi-GPUs.

    Args:
        proxy (Proxy): The NCCL communicators.
        precision (str): The precision for the state vector, single precision means complex64,
            double precision means complex128.
        gpu_device_id (int): The GPU device ID.
        sync (bool): Sync mode or Async mode.
    """
    @property
    def vector(self):
        return self.simulator.vector

    def __init__(
        self,
        proxy: Proxy,
        device: str = "GPU",
        gpu_id: int = 0,
        **options
    ):
        self.proxy = proxy
        self._data_switcher = DataSwitcher(self.proxy)
        self.simulator = CircuitSimulator() if device == "CPU" else \
            ConstantStateVectorSimulator(gpu_device_id=gpu_id, **options)

    def run(
        self,
        circuit: Circuit
    ) -> np.ndarray:
        """ start simulator with given circuit

        Args:
            circuit (Circuit): The quantum circuits.
            use_previous (bool, optional): Using the previous state vector. Defaults to False.

        Returns:
            [array]: The state vector.
        """
        self.simulator.initial_circuit(circuit)
        # initial state vector in simulator
        self.simulator.initial_state_vector(all_zeros=(self._data_switcher.id != 0))
        self._pipeline = circuit.gates
        while len(self._pipeline) > 0:
            op = self._pipeline.pop(0)

            if isinstance(op, BasicGate):
                self.simulator.apply_gate(op)
            elif isinstance(op, DeviceTrigger):
                related_gate = op.mapping(self._data_switcher.id)
                if self.simulator._precision == np.complex64:
                    related_gate.convert_precision()

                self._pipeline = related_gate.gates + self._pipeline
            elif isinstance(op, DataSwitch):
                self._data_switcher(op, self.vector)
            elif isinstance(op, Multiply):
                self.simulator.apply_multiply(op.value)
            elif isinstance(op, SpecialGate):
                self.apply_specialgate(op)
            else:
                raise TypeError("Unsupportted operator in Multi-Nodes Simulator.")

        return self.simulator.vector

    def apply_specialgate(self, op: SpecialGate):
        """ 1. Get prob
                a. if divided, using 0 or all
                b. if normal, calculate normal
        Args:
            op (SpecialGate): _description_
        """
        if op.proxy_idx != -1:
            if self._data_switcher.id & op.proxy_idx:
                temp = cp.zeros(1, dtype=np.complex128)
                prob = temp[0].real
            else:
                # prob calculation for all switch
                prob = mn_measureprob_calculator(self.vector).real
        else:
            prob = self.simulator.get_measured_prob(op.targ)

        total_prob = self._data_switcher.add_prob(prob)
        total_prob = total_prob.get()

        if op.proxy_idx != -1:
            if op.type == GateType.measure:
                self._apply_measure(op.proxy_idx, total_prob)
            elif op.type == GateType.reset:
                self._apply_reset(op.proxy_idx, total_prob)
        else:
            self.simulator.apply_specialgate(self.simulator._qubits - 1 - op.targ, op.type, total_prob)

    def _apply_measure(self, proxy_idx: int, prob: float):
        """ The algorithm for the Measure gate.

        Args:
            index (int): The target qubit of the applied quantum gate.

        Returns:
            [bool]: state 0 or state 1
        """
        _0 = random.random()        # random mistake
        if _0 > prob:  # result in state 1
            alpha = np.float32(1 / np.sqrt(1 - prob)) if self.simulator._precision == np.complex64 else \
                np.float64(1 / np.sqrt(1 - prob))

            if self._data_switcher.id & proxy_idx:
                self.simulator.apply_multiply(alpha)
            else:
                self.simulator.apply_zeros()
        else:       # result in state 0
            alpha = np.float32(1 / np.sqrt(prob)) if self.simulator._precision == np.complex64 else \
                np.float64(1 / np.sqrt(prob))

            if self._data_switcher.id & proxy_idx:
                self.simulator.apply_zeros()
            else:
                self.simulator.apply_multiply(alpha)

        return _0 > prob

    def _apply_reset(self, proxy_idx: int, prob: float):
        """ The algorithm for the Reset gate.

        Args:
            index (int): The target qubit of the applied quantum gate.
        """
        alpha = np.float64(np.sqrt(prob))
        if alpha < 1e-6:
            destination = self._data_switcher.id ^ proxy_idx
            if not (self._data_switcher.id & proxy_idx):
                self.simulator.apply_zeros()
                self._data_switcher.all_switch(self.vector, destination)
        else:
            if self._data_switcher.id & proxy_idx:
                self.simulator.apply_zeros()
            else:
                self.simulator.apply_multiply(np.float64(1 / alpha))



#!/usr/bin/env python3

import asyncio as aio
from threading import Thread
from time import time

from torch.utils.tensorboard import SummaryWriter

from QuICT.qcda.mapping.ai.data_def import TrainConfig
from QuICT.qcda.mapping.ai.train.actor import Actor
from QuICT.qcda.mapping.ai.train.learner import Learner
from QuICT.tools.logger import Logger


logger = Logger("rl-mapping-trainer")


class Trainer:
    def __init__(self, config: TrainConfig) -> None:
        logger.info("Initializing trainer...")

        # Copy values in.
        self.config = config
        self.learner = Learner(config=config)
        self.actor = Actor(rank=1, config=config)
        self.actor.policy_net = self.learner._policy_net

        self._writer = SummaryWriter(log_dir=config.log_dir)

        self._loop = aio.new_event_loop()
        self._t = Thread(target=self.side_thread, daemon=True)
        self._t.start()

    def side_thread(self):
        aio.set_event_loop(self._loop)
        self._loop.run_forever()

    async def write_stat(self, running_loss: float, running_reward: float, g_step: int):
        self._writer.add_scalar(
            tag="loss",
            scalar_value=running_loss,
            global_step=g_step,
        )
        self._writer.add_scalar(
            tag="reward",
            scalar_value=running_reward,
            global_step=g_step,
        )

    def train(self):
        running_loss = 0.0
        running_reward = 0.0
        last_obs_time = time()
        observe_period = 100
        logger.info(f"Training on {self.config.device}...\n")
        g_step = 0
        for epoch_id in range(self.config.total_epoch):
            logger.info(f"Epoch {epoch_id}")

            self.actor.agent.reset_explore_state()
            for it in range(self.config.explore_period):
                reward = self.actor.explore()

                loss = self.learner.optimize_model()

                if loss is not None:
                    running_loss += loss
                    running_reward += reward

                    aio.run_coroutine_threadsafe(
                        self.write_stat(loss, reward, g_step),
                        self._loop,
                    )

                    if (it + 1) % observe_period == 0:
                        running_loss /= observe_period
                        running_reward /= observe_period
                        act_rate = observe_period / (time() - last_obs_time)
                        learn_rate = self.config.batch_size * act_rate

                        logger.info(
                            f"[{str(it+1):4s}] loss: {running_loss:0.4f}, reward: {running_reward:0.2f}\n"
                            + f"       actor : {act_rate:0.2f} action/s, learner: {learn_rate:0.2f} transition/s"
                        )

                        running_loss = 0.0
                        running_reward = 0.0
                        last_obs_time = time()

                # transfer replay buffer
                for transition in self.actor.replay:
                    self.learner.replay.push(transition)
                self.actor.replay.clear()

                # sync policy net
                self.actor.policy_net = self.learner._policy_net

                if g_step % self.config.target_update_period == 0:
                    self.learner._target_net.load_state_dict(
                        self.learner._policy_net.state_dict()
                    )
                g_step += 1

            # validate
            v_results = self.learner.validate_model()
            self.learner.show_validation_results(v_results, epoch_id)


if __name__ == "__main__":
    import sys

    assert len(sys.argv) > 1
    topo = sys.argv[1]
    device = "cuda" if len(sys.argv) == 2 else sys.argv[2]
    config = TrainConfig(topo=topo, device=device)
    trainer = Trainer(config=config)
    trainer.train()



from ._operator import Operator
from .trigger import Trigger
from .data_switch import DataSwitch, DataSwitchType
from .device_trigger import DeviceTrigger
from .special import Multiply, SpecialGate
from .check_point import CheckPoint, CheckPointChild
from .noise_gate import NoiseGate



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/2/23 10:14 上午
# @Author  : Han Yu
# @File    : search_with_prior_knowledge.py

import numpy as np
from scipy.optimize import minimize

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.core.gate.backend import MCTOneAux
from QuICT.qcda.synthesis.quantum_state_preparation import QuantumStatePreparation

P_GLOBAL = []
T_GLOBAL = 1


def fun(x):
    return -np.dot(P_GLOBAL, np.sin((2 * T_GLOBAL + 1) * np.arcsin(np.sqrt(x)))**2)


def run_search_with_prior_knowledge(f, n, p, T, oracle):
    """ grover search for f with custom oracle

    Args:
        f(list<int>): the function to be decided
        n(int): the length of input
        p(list<float>): the distribute of priop probability
        T(int): the query times
        oracle(function): the oracle
    Returns:
        int: the a satisfies that f(a) = 1
    """
    global P_GLOBAL
    P_GLOBAL = p[:]
    global T_GLOBAL
    T_GLOBAL = T
    num = int(np.ceil(np.log2(n))) + 2
    # Determine number of qreg
    circuit = Circuit(num)
    qreg = circuit([i for i in range(num - 2)])
    ancilla = circuit(num - 2)
    empty = circuit(num - 1)
    cons = ({'type': 'ineq', 'fun': lambda x: 1 - np.dot(np.ones(n), x)},)
    bnd = [(0, np.sin(np.pi / (4 * T + 2))**2) for _ in range(n)]
    tmp = np.min([1 / n, np.sin(np.pi / (4 * T + 2))**2])
    x = np.array([tmp] * n)
    option = {'maxiter': 4, 'disp': False}
    res = minimize(fun, x, method='SLSQP', constraints=cons,
                   bounds=bnd, options=option)
    q = res.x

    # Start with qreg in equal superposition and ancilla in |->
    QSP = QuantumStatePreparation('uniformly_gates')
    gates_preparation = QSP.execute(list(q))

    MCTOA = MCTOneAux()
    gates_mct = MCTOA.execute(num)

    X | ancilla
    H | ancilla
    gates_preparation | qreg
    for i in range(T):
        oracle(f, qreg, ancilla)
        gates_preparation ^ qreg
        X | qreg
        gates_mct | circuit
        X | qreg
        gates_preparation | qreg
    # Apply H
    H | ancilla
    X | ancilla
    # Measure
    Measure | qreg
    Measure | ancilla
    Measure | empty
    circuit.exec()
    return int(qreg)


class GroverWithPriorKnowledge:
    """ grover search with prior knowledge

    https://arxiv.org/abs/2009.08721

    """
    @staticmethod
    def run(f, n, p, T, oracle):
        """ grover search for f with custom oracle

        Args:
            f(list<int>): the function to be decided
            n(int): the length of input
            p(list<float>): the distribute of priop probability
            T(int): the query times
            oracle(function): the oracle
        Returns:
            int: the a satisfies that f(a) = 1
        """
        return run_search_with_prior_knowledge(f, n, p, T, oracle)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/15 10:31
# @Author  : Han Yu, Li Kaiqi
# @File    : _circuit_computing.py
from collections import Iterable
from typing import List
import numpy as np
from enum import Enum

from .gate_type import GateType
from .circuit_matrix import CircuitMatrix, get_gates_order_by_depth


class CircuitBased(object):
    """ Based Class for Circuit and Composite Gate. """
    @property
    def name(self) -> int:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def gates(self) -> list:
        return self._gates

    @gates.setter
    def gates(self, gates):
        self._gates = gates

    def __init__(self, name: str):
        self._name = name
        self._gates = []
        self._gate_type = {}        # gate_type: # of gates
        self._pointer = None
        self._precision = np.complex128

    def size(self) -> int:
        """ the number of gates in the circuit/CompositeGate

        Returns:
            int: the number of gates in circuit
        """
        return len(self._gates)

    def width(self):
        """ the number of qubits in circuit

        Returns:
            int: the number of qubits in circuit
        """
        return len(self.qubits)

    def depth(self) -> int:
        """ the depth of the circuit/CompositeGate.

        Returns:
            int: the depth
        """
        depth = np.zeros(self.width(), dtype=int)
        for gate in self._gates:
            targs = gate.cargs + gate.targs
            depth[targs] = np.max(depth[targs]) + 1

        return np.max(depth)

    def count_2qubit_gate(self) -> int:
        """ the number of the two qubit gates in the circuit/CompositeGate

        Returns:
            int: the number of the two qubit gates
        """
        count = 0
        for gate in self._gates:
            if gate.controls + gate.targets == 2:
                count += 1

        return count

    def count_1qubit_gate(self) -> int:
        """ the number of the one qubit gates in the circuit/CompositeGate

        Returns:
            int: the number of the one qubit gates
        """
        count = 0
        for gate in self._gates:
            if gate.is_single():
                count += 1

        return count

    def count_gate_by_gatetype(self, gate_type: GateType) -> int:
        """ the number of the gates which are some type in the circuit/CompositeGate

        Args:
            gateType(GateType): the type of gates to be count

        Returns:
            int: the number of the gates which are some type
        """
        if gate_type in self._gate_type.keys():
            return self._gate_type[gate_type]

        return 0

    def __str__(self):
        circuit_info = {
            "name": self.name,
            "width": self.width(),
            "size": self.size(),
            "depth": self.depth(),
            "1-qubit gates": self.count_1qubit_gate(),
            "2-qubit gates": self.count_2qubit_gate()
        }

        return str(circuit_info)

    def qasm(self, output_file: str = None):
        """ The qasm of current CompositeGate/Circuit.

        Args:
            output_file (str): The output qasm file's name or path

        Returns:
            str: The string of Circuit/CompositeGate's qasm.
        """
        qreg = self.width()
        creg = min(self.count_gate_by_gatetype(GateType.measure), qreg)
        if creg == 0:
            creg = qreg

        qasm_string = 'OPENQASM 2.0;\ninclude "qelib1.inc";\n'
        qasm_string += f"qreg q[{qreg}];\n"
        qasm_string += f"creg c[{creg}];\n"

        cbits = 0
        for gate in self._gates:
            if gate.qasm_name == "measure":
                qasm_string += f"measure q[{gate.targ}] -> c[{cbits}];\n"
                cbits += 1
                cbits = cbits % creg
            else:
                qasm_string += gate.qasm()

        if output_file is not None:
            with open(output_file, 'w+') as of:
                of.write(qasm_string)

        return qasm_string

    def get_lastcall_for_each_qubits(self) -> List[GateType]:
        lastcall_per_qubits = [None] * self.width()
        inside_qargs = []
        for i in range(self.size() - 1, -1, -1):
            gate_args = self._gates[i].cargs + self._gates[i].targs
            gate_type = self._gates[i].type
            for garg in gate_args:
                if lastcall_per_qubits[garg] is None:
                    lastcall_per_qubits[garg] = gate_type
                    inside_qargs.append(garg)

            if len(inside_qargs) == self.width():
                break

        return lastcall_per_qubits

    def get_gates_order_by_depth(self) -> List[List]:
        """ Order the gates of circuit by its depth layer

        Returns:
            List[List[BasicGate]]: The list of gates which at same layers in circuit.
        """
        return get_gates_order_by_depth(self.gates)

    def matrix(self, device: str = "CPU", mini_arg: int = 0) -> np.ndarray:
        """ Generate the circuit's unitary matrix which compose by all quantum gates' matrix in current circuit.

        Args:
            device (str, optional): The device type for generate circuit's matrix, one of [CPU, GPU]. Defaults to "CPU".
            mini_arg (int, optional): The minimal qubit args, only use for CompositeGate local mode. Default to 0.
        """
        assert device in ["CPU", "GPU"]
        circuit_matrix = CircuitMatrix(device)

        if self.size() == 0:
            if device == "CPU":
                circuit_matrix = np.identity(1 << self.width(), dtype=self._precision)
            else:
                import cupy as cp

                circuit_matrix = cp.identity(1 << self.width(), dtype=self._precision)

            return circuit_matrix

        if self.size() > self.count_1qubit_gate() + self.count_2qubit_gate():
            self.gate_decomposition()

        return circuit_matrix.get_unitary_matrix(self.gates, self.width(), mini_arg)

    def convert_precision(self):
        """ Convert all gates in Cicuit/CompositeGate into single precision. """
        for gate in self.gates:
            if hasattr(gate, "convert_precision"):
                gate.convert_precision()

        self._precision = np.complex64 if self._precision == np.complex128 else np.complex128

    def gate_decomposition(self):
        added_idxes = 0     # The number of gates which add from gate.build_gate()
        for i in range(self.size()):
            gate = self.gates[i + added_idxes]
            if hasattr(gate, "build_gate"):
                decomp_gates = gate.build_gate()
                self.gates.remove(gate)
                for g in decomp_gates:
                    self._gates.insert(i + added_idxes, g)
                    added_idxes += 1

                added_idxes -= 1    # minus the original gate


class CircuitMode(Enum):
    Clifford = "Clifford"
    CliffordRz = "CliffordRz"
    Arithmetic = 'Arithmetic'
    Misc = "Misc"



import time
import multiprocessing
import random

from redis_controller import RedisController
from server.utils import (
    delete_job_folder, JobOperatorType, delete_user_folder,
    JobState, ResourceOp, user_resource_op
)


class MasterNodeManager:
    def __init__(self):
        self.check_interval = 15
        self.redis_connection = RedisController()

    def start(self) -> None:
        """Start agents."""
        pending_job_agent = PendingJobProcessor(
            redis_connection=self.redis_connection,
            check_interval=self.check_interval,
        )
        pending_job_agent.start()

        running_job_agent = RunningJobProcessor(
            redis_connection=self.redis_connection,
            check_interval=self.check_interval,
        )
        running_job_agent.start()

        operator_queue_agent = OperatorQueueProcessor(
            redis_connection=self.redis_connection,
            check_interval=self.check_interval,
        )
        operator_queue_agent.start()


class PendingJobProcessor(multiprocessing.Process):
    def __init__(self, redis_connection: RedisController, check_interval: int = 10):
        super().__init__()
        self.redis_connection = redis_connection
        self.check_interval = check_interval

    def run(self):
        while True:
            self._check_pending_jobs()
            time.sleep(self.check_interval)

    def _check_pending_jobs(self):
        # Get current pending jobs
        pending_jobs = self.redis_connection.get_pending_jobs_queue()

        for job_name in pending_jobs:
            job_detail = self.redis_connection.get_job_info(job_name)

            # User's limitation
            user_info = self.redis_connection.get_user_dynamic_info(job_detail['username'])
            is_user_satisfied, updated_user_info = user_resource_op(
                user_info, int(job_detail['circuit_info']['width']), job_detail["device"], ResourceOp.Allocation
            )
            if not is_user_satisfied:
                continue

            # Start job
            self._start_job(job_detail)
            self.redis_connection.add_running_job_from_pending_jobs(job_name)
            # Update resource
            self.redis_connection.update_user_dynamic_info(job_detail['username'], updated_user_info)

    def _start_job(self, job_detail: dict):
        pass


class RunningJobProcessor(multiprocessing.Process):
    def __init__(self, redis_connection: RedisController, check_interval: int = 10):
        super().__init__()
        self.redis_connection = redis_connection
        self.check_interval = check_interval

    def run(self):
        while True:
            self._check_running_jobs()
            time.sleep(self.check_interval)

    def _check_running_jobs(self):
        # Get running jobs
        running_jobs = self.redis_connection.get_running_jobs_queue()

        for job_name in running_jobs:
            # TODO: check running job status
            job_detail = self.redis_connection.get_job_info(job_name)
            user_info = self.redis_connection.get_user_dynamic_info(job_detail['username'])

            job_state = self._check_job_state(job_name)
            is_finish = job_state in [JobState.Finish, JobState.Error]
            if not is_finish:
                continue

            self.redis_connection.add_finish_job_from_running_jobs(job_name, job_state)
            # Release User's resource
            _, updated_user_info = user_resource_op(
                user_info, int(job_detail['circuit_info']['width']), job_detail["device"], ResourceOp.Release
            )
            self.redis_connection.update_user_dynamic_info(job_detail['username'], updated_user_info)

    def _check_job_state(self, job_name: str):
        # FIXME: temporary fake code, only test pipeline work.
        temp = random.random()
        if temp >= 0.5:
            return JobState.Finish

        if temp <= 0.1:
            return JobState.Error

        return JobState.Running


class OperatorQueueProcessor(multiprocessing.Process):
    def __init__(self, redis_connection: RedisController, check_interval: int = 10):
        super().__init__()
        self.redis_connection = redis_connection
        self.check_interval = check_interval

    def run(self):
        while True:
            self._check_operator_queue()
            time.sleep(self.check_interval)

    def _check_operator_queue(self):
        # Get Operator Queue
        operator_queue = self.redis_connection.get_operator_queue()

        for op in operator_queue:
            job_name, operator = op[:-4], op[-3:]
            job_detail = self.redis_connection.get_job_info(job_name)
            if operator == JobOperatorType.delete.value:
                is_success = self._delete_related_job(job_name, job_detail)
            else:
                is_success = self._delete_user_info(job_name)

            if is_success:
                self.redis_connection.remove_operator(op)

    def _delete_related_job(self, job_name: str, job_detail: dict):
        username = job_detail['username']
        job_state = job_detail['state']

        if job_state == JobState.Running.value:
            # TODO: kill running job

            # Get User Info
            user_info = self.redis_connection.get_user_dynamic_info(username)
            # Release User resource
            _, updated_user_resource = user_resource_op(
                user_info, int(job_detail['circuit_info']['width']), job_detail["device"], ResourceOp.Release
            )

            self.redis_connection.update_user_dynamic_info(username, updated_user_resource)

        self.redis_connection.remove_job(job_name, job_state)
        delete_job_folder(username, job_detail['job_name'])

        return True

    def _delete_user_info(self, user_name: str) -> bool:
        user_info = self.redis_connection.get_user_dynamic_info(user_name)
        if user_info["number_of_running_jobs"] == 0:
            self.redis_connection.delete_user_dynamic_info(user_name)
            delete_user_folder(user_name)

            return True

        return False


if __name__ == "__main__":
    manager = MasterNodeManager()
    manager.start()
